"""
PySCeS - Python Simulator for Cellular Systems (http://pysces.sourceforge.net)

Copyright (C) 2004-2022 B.G. Olivier, J.M. Rohwer, J.-H.S Hofmeyr all rights reserved,

Brett G. Olivier (bgoli@users.sourceforge.net)
Triple-J Group for Molecular Cell Physiology
Stellenbosch University, South Africa.

Permission to use, modify, and distribute this software is given under the
terms of the PySceS (BSD style) license. See LICENSE.txt that came with
this distribution for specifics.

NO WARRANTY IS EXPRESSED OR IMPLIED.  USE AT YOUR OWN RISK.
Brett G. Olivier
"""

from __future__ import division, print_function
from __future__ import absolute_import
from __future__ import unicode_literals

"""
This module contains simplified methods derived from the Pysces model class
Brett G. Olivier June 2010
"""

try:
    input = raw_input  # Py2 compatibility
except NameError:
    pass
import os, copy, time
import numpy
from pysces import PyscesStoich
from pysces import PyscesParse

mach_spec = numpy.MachAr()
pscParser = PyscesParse.PySCeSParser(debug=0)


class PyscesInputFileParser(object):
    """
    This class contains the PySCeS model loading and Stoichiometric Analysis methods
    """

    ModelDir = None
    ModelFile = None
    ModelOutput = None
    __settings__ = None
    N = None

    def __init__(self, model_file, directory, output_dir=None):
        self.ModelDir = directory
        self.ModelFile = model_file
        if output_dir == None:
            self.ModelOutput = os.getcwd()
        else:
            assert os.path.exists(output_dir), "\n%s is not a valid path" % output_dir
        self.__settings__ = {}
        # Initialize stoichiometric precision
        self.__settings__['stoichiometric_analysis_fp_zero'] = mach_spec.eps * 2.0e4
        self.__settings__['stoichiometric_analysis_lu_precision'] = self.__settings__[
            'stoichiometric_analysis_fp_zero'
        ]
        self.__settings__['stoichiometric_analysis_gj_precision'] = (
            self.__settings__['stoichiometric_analysis_lu_precision'] * 10.0
        )
        self.__settings__['enable_deprecated_attr'] = False
        self.InitialiseInputFile()
        self.N = self.buildN()

    def InitialiseInputFile(self):
        """
        InitialiseInputFile()

        Parse the input file associated with the PySCeS model instance and assign the basic model attributes

        Arguments:
        None

        """
        self.__parseOK = 1  # check that model has parsed ok?
        try:
            if os.path.exists(os.path.join(self.ModelDir, self.ModelFile)):
                pass
            else:
                print(
                    '\nInvalid self.ModelFile: '
                    + os.path.join(self.ModelDir, self.ModelFile)
                )
        except:
            print(
                'WARNING: Problem verifying: '
                + os.path.join(self.ModelDir, self.ModelFile)
            )

        if self.ModelFile[-4:] == '.psc':
            pass
        else:
            print('Assuming extension is .psc')
            self.ModelFile += '.psc'

        print('\nParsing file: %s' % os.path.join(self.ModelDir, self.ModelFile))

        pscParser.ParsePSC(self.ModelFile, self.ModelDir, self.ModelOutput)
        print(' ')

        badlist = pscParser.KeywordCheck(pscParser.ReactionIDs)
        badlist = pscParser.KeywordCheck(pscParser.Inits, badlist)

        if len(badlist) != 0:
            print(
                '\n******************************\nPSC input file contains PySCeS keywords please rename them and reload:'
            )
            for item in badlist:
                print('   --> ' + item)
            print('******************************\n')
            self.__parseOK = 0
            # assert len(badlist) != 0, 'Keyword error, please check input file'

        if self.__parseOK:
            # brett 2008
            self.__nDict__ = pscParser.nDict.copy()
            self.__sDict__ = pscParser.sDict.copy()
            self.__pDict__ = pscParser.pDict.copy()
            self.__uDict__ = pscParser.uDict.copy()

            # model attributes are now initialised here brett2008
            self.__InitDict__ = {}
            # set parameters and add to __InitDict__
            for p in list(self.__pDict__.keys()):
                setattr(self, self.__pDict__[p]['name'], self.__pDict__[p]['initial'])
                self.__InitDict__.update(
                    {self.__pDict__[p]['name']: self.__pDict__[p]['initial']}
                )
            # set species and add to __InitDict__ and set mod.Xi_init
            for s in list(self.__sDict__.keys()):
                setattr(self, self.__sDict__[s]['name'], self.__sDict__[s]['initial'])
                if not self.__sDict__[s]['fixed']:
                    setattr(
                        self,
                        self.__sDict__[s]['name'] + '_init',
                        self.__sDict__[s]['initial'],
                    )
                self.__InitDict__.update(
                    {self.__sDict__[s]['name']: self.__sDict__[s]['initial']}
                )

            # setup keywords
            self.__KeyWords__ = pscParser.KeyWords.copy()
            if self.__KeyWords__['Modelname'] == None:
                self.__KeyWords__['Modelname'] = self.ModelFile.replace('.psc', '')
            if self.__KeyWords__['Description'] == None:
                self.__KeyWords__['Description'] = self.ModelFile.replace('.psc', '')
            # if SpeciesTypes undefined assume []
            if self.__KeyWords__['Species_In_Conc'] == None:
                self.__KeyWords__['Species_In_Conc'] = True
            # if OutputType is undefined assume it is the same as SpeciesType
            if self.__KeyWords__['Output_In_Conc'] == None:
                if self.__KeyWords__['Species_In_Conc']:
                    self.__KeyWords__['Output_In_Conc'] = True
                else:
                    self.__KeyWords__['Output_In_Conc'] = False
            # set the species type in sDict according to 'Species_In_Conc'
            for s in list(self.__sDict__.keys()):
                if not self.__KeyWords__['Species_In_Conc']:
                    self.__sDict__[s]['isamount'] = True
                else:
                    self.__sDict__[s]['isamount'] = False

            # setup compartments
            self.__compartments__ = pscParser.compartments.copy()
            if len(list(self.__compartments__.keys())) > 0:
                self.__HAS_COMPARTMENTS__ = True
            else:
                self.__HAS_COMPARTMENTS__ = False

            # no (self.)
            self.__fixed_species__ = copy.copy(pscParser.fixed_species)
            self.__species__ = copy.copy(pscParser.species)
            self.__parameters__ = copy.copy(pscParser.parameters)
            self.__reactions__ = copy.copy(pscParser.reactions)
            self.__modifiers__ = copy.copy(pscParser.modifiers)
            # Initialize exposed stuff
            self.fixed_species = tuple(pscParser.fixed_species)
            self.species = tuple(pscParser.species)
            self.parameters = tuple(pscParser.parameters)
            self.reactions = tuple(pscParser.reactions)
            self.modifiers = tuple(pscParser.modifiers)

            # Add input file defined fuctions - brett 200500621
            # TODO deprecated
            # self._Function_time = copy.copy(pscParser.TimeFunc)
            self._Function_user = copy.copy(pscParser.UserFunc)
            self._Function_init = pscParser.InitFunc

            self.__functions__ = pscParser.Functions.copy()
            self.__rules__ = pscParser.AssignmentRules.copy()
            self.__InitFuncs__ = pscParser.ModelInit.copy()
            self.__userfuncs__ = pscParser.UserFuncs.copy()
            self.__eDict__ = pscParser.Events.copy()

            ##  if pscParser.ModelUsesNumpyFuncs:
            ##  print 'Numpy functions detected in kinetic laws.\n'
        else:
            print('\nERROR: model parsing error, please check input file.\n')
        # added in a check for model correctness and human error reporting (1=ok, 0=error)
        if len(pscParser.SymbolErrors) != 0:
            print('\nUndefined symbols:\n%s' % self.SymbolErrors)
        if not pscParser.ParseOK:
            print(
                '\n\n*****\nModel parsing errors detected in input file '
                + self.ModelFile
                + '\n*****'
            )
            print('\nInput file errors')
            for error in pscParser.LexErrors:
                print(
                    error[0]
                    + 'in line:\t'
                    + str(error[1])
                    + ' ('
                    + error[2][:20]
                    + ' ...)'
                )
            print('\nParser errors')
            for error in pscParser.ParseErrors:
                try:
                    print(error[0] + '- ' + error[2][:20])
                except:
                    print(error)
            assert pscParser.ParseOK == 1, 'Input File Error'

    def buildN(self):
        """
        buildN()

        Generate the stoichiometric matrix N from the parsed model description.
        Returns a stoichiometric matrix (N)

        Arguments:
        None

        """
        VarReagents = ['self.' + s for s in self.__species__]
        StoicMatrix = numpy.zeros((len(VarReagents), len(self.__reactions__)), 'd')
        for reag in VarReagents:
            for id in self.__reactions__:
                if reag in list(self.__nDict__[id]['Reagents'].keys()):
                    StoicMatrix[VarReagents.index(reag)][
                        self.__reactions__.index(id)
                    ] = self.__nDict__[id]['Reagents'][reag]
        return StoicMatrix

    def Stoichiometry_Init(self, nmatrix):
        """
        Stoichiometry_Init(nmatrix,load=0)

        Initialize the model stoichiometry. Given a stoichiometric matrix N, this method will return an instantiated PyscesStoich instance and status flag.
        and test it's correctness. The status flag indicates 0 = reanalyse stoichiometry or
        1 = complete structural analysis preloaded.

        Arguments:

        nmatrix: The input stoichiometric matrix, N
        load [default=0]: try to load a saved stoichiometry (1)

        """
        # print 'Instantiating new stoichiometry ...'
        stc = PyscesStoich.Stoich(nmatrix)
        status = 0
        return stc, status

    def Stoichiometry_Analyse(self):
        override = 0
        load = 0
        """
        Stoichiometry_Analyse(override=0,load=0)

        Perform a structural analyses. The default behaviour is to construct and analyse the model
        from the parsed model information. Overriding this behaviour analyses the stoichiometry
        based on the current stoichiometric matrix. If load is specified PySCeS tries to load a
        saved stoichiometry, otherwise the stoichiometric analysis is run. The results of
        the analysis are checked for floating point error and nullspace rank consistancy.

        Arguments:

        override [default=0]: override stoichiometric analysis intialisation from parsed data
        load [default=0]: load a presaved stoichiometry

        """
        if not override:
            self.nmatrix = self.buildN()  # Creates the model N
            # print '\nintializing N\n'
        else:
            print('\nStoichiometric override active\n')

        assert len(self.nmatrix) > 0, (
            '\nUnable to generate Stoichiometric Matrix! model has:\n%s reactions\n%s species\nwhat did you have in mind?\n'
            % (len(self.__reactions__), len(self.__species__))
        )

        ##  self.__nmatrix__ = copy.copy(self.nmatrix)
        self.__nmatrix__ = self.nmatrix  # done with caution brett2008
        self.__Nshape__ = self.nmatrix.shape  # Get the shape of N
        ##  self.__Vtemp__ = numpy.zeros((self.__Nshape__[1])) # going going ....

        # get stoich instance and whether it was analysed or loaded - brett 20050830
        self.__structural__, stc_load = self.Stoichiometry_Init(self.nmatrix)

        # if not loaded analyze - brett 20050830
        if not stc_load:
            # technically this means we can define this on the fly - brett #20051013
            self.__structural__.stoichiometric_analysis_fp_zero = self.__settings__[
                'stoichiometric_analysis_fp_zero'
            ]
            self.__structural__.stoichiometric_analysis_lu_precision = self.__settings__[
                'stoichiometric_analysis_lu_precision'
            ]
            self.__structural__.stoichiometric_analysis_gj_precision = self.__settings__[
                'stoichiometric_analysis_gj_precision'
            ]
            self.__structural__.AnalyseL()  # Get all L related stuff
            self.__structural__.AnalyseK()  # Get all K related stuff

        # test matrix values against __settings__['stoichiometric_analysis_lu_precision']
        lsmall, lbig = self.__structural__.MatrixValueCompare(
            self.__structural__.lzeromatrix
        )
        ksmall, kbig = self.__structural__.MatrixValueCompare(
            self.__structural__.kzeromatrix
        )
        SmallValueError = 0
        if (
            abs(lsmall)
            < self.__structural__.stoichiometric_analysis_lu_precision * 10.0
        ):
            print(
                '\nWARNING: values in L0matrix are close to stoichiometric precision!'
            )
            print(
                'Stoichiometric LU precision:',
                self.__structural__.stoichiometric_analysis_lu_precision,
            )
            print('L0 smallest abs(value)', abs(lsmall))
            print('Machine precision:', mach_spec.eps)
            SmallValueError = 1
        if (
            abs(ksmall)
            < self.__structural__.stoichiometric_analysis_lu_precision * 10.0
        ):
            print(
                '\nWARNING: values in K0matrix are close to stoichiometric precision!'
            )
            print(
                'Stoichiometric precision:',
                self.__structural__.stoichiometric_analysis_lu_precision,
            )
            print('K0 smallest abs(value)', abs(ksmall))
            print('Machine precision:', mach_spec.eps)
            SmallValueError = 1
        if SmallValueError:
            input(
                '\nStructural Analysis results may not be reliable!!!.\n\nTry change <mod>.__settings__["stoichiometric_analysis_lu_precision"] (see reference manual for details)\n\n\t press any key to continue: '
            )

        # cross check that rank is consistant between K0 and L0
        if (
            self.__structural__.kzeromatrix.shape[0]
            != self.__structural__.lzeromatrix.shape[1]
        ):
            print(
                '\nWARNING: the rank calculated by the Kand L analysis methods are not the same!'
            )
            print(
                '\tK analysis calculates the rank as: '
                + repr(self.__structural__.kzeromatrix.shape[0])
            )
            print(
                '\tL analysis calculates the rank as: '
                + repr(self.__structural__.lzeromatrix.shape[1])
            )
            print('This is not good! Structural Analysis results are not reliable!!!\n')
            assert (
                self.__structural__.kzeromatrix.shape[0]
                == self.__structural__.lzeromatrix.shape[1]
            ), '\nStructuralAnalysis Error: rank mismatch'

        self.__HAS_FLUX_CONSERVATION__ = self.__structural__.info_flux_conserve
        self.__HAS_MOIETY_CONSERVATION__ = self.__structural__.info_moiety_conserve

        if self.__settings__['enable_deprecated_attr']:
            self.nmatrix_row = self.__structural__.nmatrix_row
            self.nmatrix_col = self.__structural__.nmatrix_col

            self.kmatrix = self.__structural__.kmatrix
            self.kmatrix_row = self.__structural__.kmatrix_row
            self.kmatrix_col = self.__structural__.kmatrix_col
            self.kzeromatrix = self.__structural__.kzeromatrix
            self.kzeromatrix_row = self.__structural__.kzeromatrix_row
            self.kzeromatrix_col = self.__structural__.kzeromatrix_col

            self.lmatrix = self.__structural__.lmatrix
            self.lmatrix_row = self.__structural__.lmatrix_row
            self.lmatrix_col = self.__structural__.lmatrix_col
            self.lzeromatrix = self.__structural__.lzeromatrix
            self.lzeromatrix_row = self.__structural__.lzeromatrix_row
            self.lzeromatrix_col = self.__structural__.lzeromatrix_col
            self.conservation_matrix = self.__structural__.conservation_matrix
            self.conservation_matrix_row = self.__structural__.conservation_matrix_row
            self.conservation_matrix_col = self.__structural__.conservation_matrix_col
            self.nrmatrix = self.__structural__.nrmatrix
            self.nrmatrix_row = self.__structural__.nrmatrix_row
            self.nrmatrix_col = self.__structural__.nrmatrix_col

            self.__kmatrix__ = copy.copy(self.kmatrix)
            self.__kzeromatrix__ = copy.copy(self.kzeromatrix)
            self.__lmatrix__ = copy.copy(self.lmatrix)
            self.__lzeromatrix__ = copy.copy(self.lzeromatrix)
            self.__nrmatrix__ = copy.copy(self.nrmatrix)
        # switch that is set if the stoichiometric analysis is up to date

        self.__structural__.species = self.species
        self.__structural__.reactions = self.reactions
        self.Nmatrix = PyscesStoich.StructMatrix(
            self.__structural__.nmatrix,
            self.__structural__.nmatrix_row,
            self.__structural__.nmatrix_col,
        )
        self.Nmatrix.setRow(self.species)
        self.Nmatrix.setCol(self.reactions)

        self.Nrmatrix = PyscesStoich.StructMatrix(
            self.__structural__.nrmatrix,
            self.__structural__.nrmatrix_row,
            self.__structural__.nrmatrix_col,
        )
        self.Nrmatrix.setRow(self.species)
        self.Nrmatrix.setCol(self.reactions)

        self.Kmatrix = PyscesStoich.StructMatrix(
            self.__structural__.kmatrix,
            self.__structural__.kmatrix_row,
            self.__structural__.kmatrix_col,
        )
        self.Kmatrix.setRow(self.reactions)
        self.Kmatrix.setCol(self.reactions)

        self.K0matrix = PyscesStoich.StructMatrix(
            self.__structural__.kzeromatrix,
            self.__structural__.kzeromatrix_row,
            self.__structural__.kzeromatrix_col,
        )
        self.K0matrix.setRow(self.reactions)
        self.K0matrix.setCol(self.reactions)

        self.Lmatrix = PyscesStoich.StructMatrix(
            self.__structural__.lmatrix,
            self.__structural__.lmatrix_row,
            self.__structural__.lmatrix_col,
        )
        self.Lmatrix.setRow(self.species)
        self.Lmatrix.setCol(self.species)

        self.L0matrix = PyscesStoich.StructMatrix(
            self.__structural__.lzeromatrix,
            self.__structural__.lzeromatrix_row,
            self.__structural__.lzeromatrix_col,
        )
        self.L0matrix.setRow(self.species)
        self.L0matrix.setCol(self.species)

        if self.__structural__.info_moiety_conserve:
            self.Consmatrix = PyscesStoich.StructMatrix(
                self.__structural__.conservation_matrix,
                self.__structural__.conservation_matrix_row,
                self.__structural__.conservation_matrix_col,
            )
            self.Consmatrix.setRow(self.species)
            self.Consmatrix.setCol(self.species)
        else:
            self.Consmatrix = None
        self.__StoichOK = 1
        print(' ')


if __name__ == '__main__':
    ModelFile = 'pysces_test_linear1.psc'
    ModelDir = '/home/bgoli/Pysces/psc'
    mod = PyscesInputFileParser(ModelFile, ModelDir)
    # ~ mod.Stoichiometry_Analyse()



"""
PySCeS - Python Simulator for Cellular Systems (http://pysces.sourceforge.net)

Copyright (C) 2004-2022 B.G. Olivier, J.M. Rohwer, J.-H.S Hofmeyr all rights reserved,

Johann Rohwer (jrohwer@users.sourceforge.net)
Triple-J Group for Molecular Cell Physiology
Stellenbosch University, South Africa.

Permission to use, modify, and distribute this software is given under the
terms of the PySceS (BSD style) license. See LICENSE.txt that came with
this distribution for specifics.

NO WARRANTY IS EXPRESSED OR IMPLIED.  USE AT YOUR OWN RISK.
Johann M. Rohwer
"""
from __future__ import division, print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import sys, pickle
import multiprocessing
from pysces.PyscesParScan import Analyze
from time import sleep


if __name__ == '__main__':
    print("\nmultiprocessing script command line:")
    print(sys.argv)
    # called with
    # subprocess.call(['python', MULTISCANFILE, self._MODE_, fN])
    MODE = sys.argv[1]
    pool = multiprocessing.Pool()

    # load stuff from the pickle
    with open(sys.argv[2], 'rb') as F:
        mod, scanpartition, seqpartition, genorder, useroutputlist = pickle.load(F)

    mod.SetQuiet()  # kill verbose output during scan
    # append tasks to asynchronous results list
    arl = []
    for i in range(len(scanpartition)):
        arl.append(
            pool.apply_async(
                Analyze,
                (
                    scanpartition[i],
                    seqpartition[i],
                    genorder,
                    MODE,
                    useroutputlist,
                    mod,
                ),
            )
        )

    print("Submitted tasks:", len(arl))
    print("\nPARALLEL COMPUTATION\n--------------------")

    while arl[-1].ready() != True:
        sleep(5)
        print('Tasks to go... ', [r.ready() for r in arl].count(False))
        # wait until all tasks are completed
    arl[-1].wait()

    # get results
    print("\nGATHER RESULT\n-------------")
    res_list = []
    for ar in arl:
        res_list.append(ar.get())
    pool.close()
    # pickle results_list
    with open(sys.argv[2], 'wb') as F:
        pickle.dump(res_list, F, protocol=-1)
        F.flush()



"""
PySCeS - Python Simulator for Cellular Systems (http://pysces.sourceforge.net)

Copyright (C) 2004-2022 B.G. Olivier, J.M. Rohwer, J.-H.S Hofmeyr all rights reserved,

Brett G. Olivier (bgoli@users.sourceforge.net)
Triple-J Group for Molecular Cell Physiology
Stellenbosch University, South Africa.

Permission to use, modify, and distribute this software is given under the
terms of the PySceS (BSD style) license. See LICENSE.txt that came with
this distribution for specifics.

NO WARRANTY IS EXPRESSED OR IMPLIED.  USE AT YOUR OWN RISK.
Brett G. Olivier
"""
from __future__ import division, print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from pysces.version import __version__

__doc__ = "PySCeS JWS parser module -- uses  PLY 1.5 or newer"

try:
    input = raw_input  # Py2 compatibility
except NameError:
    pass

import os, copy
from .lib import lex
from .lib import yacc
from getpass import getuser
from time import sleep, strftime
from numpy import MachAr

MyMachArr = MachAr()


class JWSParser:
    """JWSParser written by Johann, based on Jannie's lexparse and integrated into PySCeS by brett  -- converts PySCeS (.psc) files to JWS Online (jws) files"""

    ReactionIDs = []  # List of reaction names
    Names = []  # List of all reagent, parameter and function names
    LexErrors = []  # List of lexing errors

    NetworkDict = {}  # Dictionary containing all reaction information
    InitStrings = []  # Initialisation strings
    InitParStrings = []  # Initialisation strings for parameters -- johann new
    InitVarStrings = []  # Initialisation strings for variables -- johann new
    Inits = []  # Initialised entities
    Reagents = []  # All reagents found during parsing of reactions
    VarReagents = []  # Variable reagents that occur in reactions
    FixedReagents = []  # Fixed reagents
    ReacParams = []  # Temporary list of reaction parameters
    InitParams = []  # Initialised parameters
    ParseErrors = []

    mach_spec = MyMachArr
    AllRateEqsGiven = 1  # Flag to check that all rate equations have been given
    Debug = 0

    ##############
    # Build the lexer
    ##############

    # elementary regular expressions used as building blocks
    Int = r'\d+'  # Integer
    Dec = Int + '\.' + Int  # Decimal

    # List of token names
    tokens = (
        'FIXDEC',
        'IRREV',
        #'REAL', # johann -- now build up real in a p function since we want to make exponent explicit
        'INT',
        'DEC',  # johann -- added explicitly since we no longer have REAL token
        'PLUS',
        'MINUS',
        'TIMES',
        'DIVIDE',
        'POWER',
        'LPAREN',
        'RPAREN',
        'EQUALS',
        'COMMA',
        'REACTION_ID',
        'STOICH_COEF',
        'NAME',
        'EXP',
    )  # johann -- new EXP token

    # Simple tokens
    t_IRREV = r'>'
    # t_REAL = Real # johann -- no longer have a real token, now a p function
    t_INT = Int
    t_DEC = Dec  # new DEC token
    t_PLUS = r'\+'
    t_MINUS = r'-'
    t_TIMES = r'\*'
    t_DIVIDE = r'/'
    t_POWER = '\*\*'
    t_LPAREN = r'\('
    t_RPAREN = r'\)'
    t_EQUALS = r'='
    t_COMMA = r','
    t_ignore = ' \t\r'  # Ignore spaces and tabs --- and windows return - brett 20040229

    def t_comment(self, t):
        r'\#.+\n'  # Match from # to newline
        t.lineno += 1  # Increment line number

    def t_newline(self, t):
        r'\n+'  # Match newline
        t.lineno += len(t.value)  # Increment with number of consecutive newlines

    def t_EXP(self, t):  # johann -- need separate EXP token to replace for Mathematica
        r'\d+\.?\d*[E|e][\+|\-]?'  # define EXP token merely as digits[.]digits[E|e][+|-]
        t.type = 'EXP'  # parse final integer separately in 'Real' p-function to remove leading zeros
        t.value = t.value.replace('e', ' 10^')
        t.value = t.value.replace('E', ' 10^')
        return t

    def t_FIXDEC(self, t):
        r'FIX:'
        t.type = 'FIXDEC'
        t.value = 'FIX:'
        return t

    def t_REACTION_ID(self, t):
        r'[a-zA-Z]\w*:'  # Match any letter followed by zero or more characters
        # in [a-zA-Z0-9_] up to a colon
        t.type = 'REACTION_ID'
        if t.value[0] == 'v' and len(t.value) > 1:
            t.value = t.value[
                1:
            ]  # remove initial 'v' if present to avoid constructions like 'v[vR1]'
        t.value = (
            'v[' + t.value[:-1] + ']'
        )  # remove the colon and add v[] for JWS -- johann
        if t.value in self.ReactionIDs:
            self.LexErrors.append(('Duplicate ReactionID ', t.lineno, t.value, t.type))
        else:
            self.ReactionIDs.append(t.value)
        return t

    def t_STOICH_COEF(self, t):
        r'\{\d+\}|\{\d+\.\d+\}'
        t.type = 'STOICH_COEF'
        t.value = t.value[1:-1]
        return t

    def t_NAME(self, t):
        r'[a-zA-Z][\w]*'  # Match any letter followed by zero or characters in the set [a-zA-Z0-9_]
        if (t.value + '[t]' not in self.Names) and (
            t.value not in self.FuncNames
        ):  # Only add to list if absent in list
            # self.Names.append(t.value)
            self.Names.append(t.value + '[t]')  # -- johann
            # print self.Names[-1]
        # hack! - brett
        if (
            t.value not in self.FuncNames
        ):  # make class attributes, ignore function names
            # print 't value before', t.value
            gt = t.value + '[t]'
            t.value = gt
        # print 't value after', t.value
        t.type = 'NAME'
        return t

    def t_error(self, t):
        self.LexErrors.append(('Lexer error ', t.lineno, t.value, t.type))
        print('Illegal character, Line ' + str(t.lineno) + ' :' + str(t.value[0]))
        t.skip(1)

    ##############
    # The parser #
    ##############

    FuncNames = (
        'acos',
        'asin',
        'atan',
        'atan2',
        'ceil',
        'cos',
        'cosh',
        'exp',
        'fabs',
        'floor',
        'fmod',
        'frexp',
        'hypot',
        'ldexp',
        'log',
        'log10',
        'modf',
        'pow',
        'sin',
        'sinh',
        'sqrt',
        'tan',
        'tanh',
    )

    precedence = (
        ('left', 'PLUS', 'MINUS'),
        ('left', 'TIMES', 'DIVIDE'),
        ('left', 'POWER'),
        ('right', 'UMINUS'),
    )

    def Show(self, name, tok):
        if self.Debug:
            print(name, tok)

    def p_error(self, t):
        self.ParseErrors.append(('Syntax error ', t.lineno, t.value, t.type))
        print('Syntax error, Line ' + str(t.lineno) + ' : ' + str(t.value))
        tok = yacc.token()
        while tok and tok.type != 'REACTION_ID':
            tok = yacc.token()
        return tok

    def p_model(self, t):
        '''Model : Statement
                 | Model Statement '''

        self.Show('Model', t[0])

    def p_statement(self, t):
        '''Statement : Fixed
                     | ReactionLine
                     | Initialise'''
        self.Show('Statement', t[0])

    def p_fixed(self, t):
        '''Fixed : FIXDEC FixList'''

        self.Show('Fixed:', t[0])

    def p_fixedreagents(self, t):
        '''FixList : NAME
                   | NAME FixList'''
        if t[1] != None:
            self.FixedReagents.append(t[1][:-3])  # johann -- remove [t] off end
        t[0] = [t[1]]
        try:
            t[0] += t[2]
        except:
            pass
        self.Show('FixList', t[0])

    def p_initialise(self, t):
        '''Initialise : NAME EQUALS Expression'''
        t[1] = t[1][:-3] + '[0]'  # johann 20050302 -- Mathematica initialisation
        t[0] = t[1] + t[2] + t[3]
        ##    del temp
        self.InitStrings.append(t[0].replace('=', ' = '))
        self.Inits.append(t[1])
        self.Show('Initialisation', t[0])

    def p_reaction_line(self, t):
        '''ReactionLine : REACTION_ID ReactionEq
                        | REACTION_ID ReactionEq Expression'''

        # global self.AllRateEqsGiven, ReacParams
        ReacID = t[1]
        if ReacID in self.NetworkDict:
            self.ParseErrors.append(('Duplicate Reaction ', t.lineno, ReacID, None))
        self.NetworkDict[ReacID] = {}  # Reaction dictionary for ReacID
        self.NetworkDict[ReacID]['Reagents'] = {}  # Reagent dictionary within ReacID

        # brett: if an index exists sum the coefficients instead of adding a new one
        # this seems to deal with multiple definitions like X + X > Y and 2{X} + Y > Z + X
        for i in t[2][
            0
        ]:  # First tuple member of ReactionEq contains list of (name,stoichcoef)
            if i[0] in self.NetworkDict[ReacID]['Reagents']:
                self.NetworkDict[ReacID]['Reagents'][i[0]] = (
                    self.NetworkDict[ReacID]['Reagents'][i[0]] + i[1]
                )
            else:
                self.NetworkDict[ReacID]['Reagents'][i[0]] = i[
                    1
                ]  # Key for reagent with stoichcoef value
        killList = []
        # brett: however for the case of X + Y > Y + Z where the sum of the coefficients
        # is zero we can delete the key (Y) out of the reaction list altgether (hopefully!)
        for i in self.NetworkDict[ReacID]['Reagents']:
            if (
                abs(self.NetworkDict[ReacID]['Reagents'][i])
                < self.mach_spec.eps * 100.0
            ):
                killList.append(i)
                # print self.mach_spec.eps*100.0, self.NetworkDict[ReacID]['Reagents']
        # print killList, self.NetworkDict[ReacID]['Reagents']
        # brett: and the easiest way of doing this is putting the zero keys in a list
        # and deleting them out of the dictionary
        if len(killList) != 0:
            for i in killList:
                del self.NetworkDict[ReacID]['Reagents'][i]
        # print killList, self.NetworkDict[ReacID]['Reagents']

        self.NetworkDict[ReacID]['Type'] = t[2][
            1
        ]  # Second tuple member of ReactionEq contains type
        try:  # Save rate equation and create parameter list
            self.NetworkDict[ReacID]['RateEq'] = t[3]
            self.NetworkDict[ReacID]['Params'] = self.ReacParams
            self.ReacParams = []  # Reset global self.ReacParams list
        except:
            self.NetworkDict[ReacID]['RateEq'] = ''
            self.NetworkDict[ReacID]['Params'] = []
            self.AllRateEqsGiven = 0  # Set global flag to false
        self.Show('ReactionLine', t[0])
        self.Show('t1', t[1])
        self.Show('t2', t[2])
        self.Show('t3', t[3])

    def p_reaction_eq(self, t):
        '''ReactionEq : LeftHalfReaction EQUALS RightHalfReaction
                      | LeftHalfReaction IRREV  RightHalfReaction'''

        ReacType = ''
        if t[2] == '=':
            ReacType = 'Rever'
        elif t[2] == '>':
            ReacType = 'Irrev'
        t[0] = (t[1] + t[3], ReacType)
        self.Show('ReactionEq', t[0])

    def p_left_half_reaction(self, t):
        ''' LeftHalfReaction : SubstrateTerm
                             | SubstrateTerm PLUS LeftHalfReaction'''

        # Make a list of substrate terms
        t[0] = [t[1]]

        try:
            t[0] += t[3]
        except:
            pass
        # brett
        # print "lhr ", t[0]
        self.Show('LeftHalfReaction', t[0])

    def p_right_half_reaction(self, t):
        ''' RightHalfReaction : ProductTerm
                              | ProductTerm PLUS RightHalfReaction'''

        # Make a list of product terms
        t[0] = [t[1]]
        try:
            t[0] += t[3]
        except:
            pass
        # brett
        # print "rhr ", t[0]
        self.Show('RightHalfReaction', t[0])

    def p_substrate_term(self, t):
        '''SubstrateTerm : STOICH_COEF NAME
                         | NAME'''

        # Make tuple of NAME and stoichiometric coefficient
        # (< 0 because substrate)
        try:
            t[0] = (t[2], -float(t[1]))
            if t[2] not in self.Reagents:
                self.Reagents.append(t[2])
        except:
            t[0] = (t[1], -1.0)
            if t[1] not in self.Reagents:
                self.Reagents.append(t[1])
        self.Show('SubstrateTerm', t[0])

    def p_product_term(self, t):
        '''ProductTerm : STOICH_COEF NAME
                       | NAME'''

        # Make tuple of NAME and stoichiometric coefficient
        # (> 0 because product)
        try:
            t[0] = (t[2], float(t[1]))
            if t[2] not in self.Reagents:
                self.Reagents.append(t[2])
        except:
            t[0] = (t[1], 1.0)
            if t[1] not in self.Reagents:
                self.Reagents.append(t[1])
        self.Show('ProductTerm', t[0])

    def p_rate_eq(self, t):
        '''Expression : Expression PLUS Expression
                      | Expression MINUS Expression
                      | Expression TIMES Expression
                      | Expression DIVIDE Expression
                      | Power
                      | Number
                      | Func'''
        # |UMINUS : add if the
        #  alternative for p_uminus is used

        if len(t.slice) == 4:
            t[0] = t[1] + t[2] + t[3]
        else:
            t[0] = t[1]

    def p_power(self, t):
        '''Power : Expression POWER Expression'''

        t[0] = (
            'Power[' + t[1] + ',' + t[3] + ']'
        )  # changed to Mathematica notation -- johann

    def p_uminus(self, t):
        '''Expression : MINUS Expression %prec UMINUS'''
        # Alternative '''UMINUS : MINUS Expression'''

        t[0] = t[1] + t[2]

    def p_number(self, t):
        '''Number : Real
                  | INT
                  | DEC
                  | NAME'''

        # Build list of entities
        try:
            float(t[1])  # check for a number
        except:
            if (
                (t[1] not in self.FuncNames)
                and (t[1] not in self.ReacParams)
                and (' 10^' not in t[1])
            ):
                # ignore function names, duplications and exponentials
                self.ReacParams.append(t[1])
                # self.ReacParams.append('self.' + t[1])
        t[0] = t[1]

    def p_real(self, t):
        '''Real : EXP INT'''
        loop = 1
        while loop == 1:  # remove leading zeros from exponent
            if t[2][0] == '0' and len(t[2]) > 1:
                t[2] = t[2][1:]
            else:
                loop = 0
        t[0] = t[1] + t[2]

    def p_function(self, t):
        '''Func : LPAREN ArgList RPAREN
                | NAME LPAREN ArgList RPAREN'''

        try:
            t[0] = t[1] + t[2] + t[3] + t[4]
        except:
            t[0] = t[1] + t[2] + t[3]

    def p_arglist(self, t):
        '''ArgList : Expression
                   | Expression COMMA Expression'''

        t[0] = t[1]
        try:
            t[0] += t[2] + t[3]
        except:
            pass

    ############################################
    # end of lexer and parser definitions
    ############################################

    def psc2jws(self, File, indir=None, outdir=None, quiet=1, debug=0):
        """
        psc2jws(File,indir=None,outdir=None,quiet=1,debug=0)
        Convert a PySCeS (.psc) file to a JWS Online (.jws) file. Call with the input file name, note the input (indir) and output (outdir) can optionally be specified.

        Arguments:
        =========
        File: PSC input file
        indir [default=None]: directory of PSC file
        outdir [default=None]: output directory for JWS file
        quiet [default=1]: turn lex/parse noise on/off
        debug [default=0]:  optionally output debug information

        """
        if indir == None:
            indir = os.getcwd()
        if outdir == None:
            outdir = os.getcwd()
        if os.path.exists(os.path.join(indir, File)) and File[-4:] == '.psc':
            go = 1
        else:
            print('\nIgnoring non-PySCeS model file: ' + os.path.join(indir, File))
            go = 0

        if go == 1:
            # clean up the modules
            reload(lex)  # brett's bugbear code these have to be here ALWAYS!!
            reload(yacc)
            # clean up the instance
            self.ReactionIDs = []  # List of reaction names
            self.Names = []  # List of all reagent, parameter and function names
            self.LexErrors = []  # List of lexing errors
            self.NetworkDict = {}  # Dictionary containing all reaction information
            self.InitStrings = []  # Initialisation strings

            self.Inits = []  # Initialised entities
            self.Reagents = []  # All reagents found during parsing of reactions
            self.FixedReagents = []  # Fixed reagents
            self.ReacParams = []  # Temporary list of reaction parameters
            self.ParseErrors = []

            self.InitParStrings = (
                []
            )  # Initialisation strings for parameters -- johann new
            self.InitVarStrings = (
                []
            )  # Initialisation strings for variables -- johann new
            self.VarReagents = []  # Variable reagents that occur in reactions
            self.InitParams = []  # Initialised parameters

            print('\nParsing file: ' + os.path.join(indir, File))

            Data = open(os.path.join(indir, File), 'r')
            Model = Data.read()
            Data.close()

            self.Debug = debug
            self.AllRateEqsGiven = (
                1  # Flag to check that all rate equations have been given
            )

            # try and find a temporary workspace or use cwd
            if 'TMP' in os.environ:
                tempDir = os.environ['TMP']
            elif 'TEMP' in os.environ:
                tempDir = os.environ['TEMP']
            else:
                tempDir = os.getcwd()

            os.chdir(tempDir)

            # fix filenames for intermediary files - brett
            if not File[:-4].isalnum():
                FileL = list(File)
                FileT = ''
                for let in FileL:
                    if let.isalnum():
                        FileT += let

                # instantiate the lexer and parser
                self.debugfile = '_jws' + FileT[:-3] + ".dbg"
                self.tabmodule = '_jws' + FileT[:-3] + "_" + "parsetab"
            else:
                self.debugfile = '_jws' + File[:-4] + ".dbg"
                self.tabmodule = '_jws' + File[:-4] + "_" + "parsetab"

            if self.Debug:
                print(self.tabmodule)
                print(self.debugfile)

            lex.lex(module=self, debug=self.Debug)
            lex.input(Model)
            yacc.yacc(
                module=self,
                debug=self.Debug,
                debugfile=self.debugfile,
                tabmodule=self.tabmodule,
            )

            os.chdir(outdir)

            while 1:
                tok = lex.token()
                if not tok:
                    break
            if self.LexErrors != []:
                print('self.LexErrors = ', self.LexErrors, '\n')

            while 1:
                p = yacc.parse(Model)
                if not p:
                    break

            # we have the dictionary get rid of this stuff
            del Model, p

            # Create list of variable reagents and remove '[t]' from fixed reagents
            for i in range(
                len(self.Reagents)
            ):  # johann -- new construction otherwise list elements not replaced
                if self.Reagents[i][:-3] not in self.FixedReagents:
                    self.VarReagents.append(self.Reagents[i])
                if self.Reagents[i][:-3] in self.FixedReagents:
                    self.Reagents[i] = self.Reagents[i][:-3]

            # Create list of initialised parameters
            for i in range(len(self.Inits)):  # johann -- reworked extensively
                if self.Inits[i][:-3] + '[t]' not in self.VarReagents:
                    self.InitStrings[i] = self.InitStrings[i].replace('[0]', '')
                    self.InitStrings[i] = self.InitStrings[i].replace(
                        '[t]', ''
                    )  # capture params initialised i.t.o. other params
                    self.Inits[i] = self.Inits[i][:-3]
                    self.InitParams.append(self.Inits[i])
                    self.InitParStrings.append(self.InitStrings[i])
                elif self.Inits[i][:-3] + '[t]' in self.VarReagents:
                    self.InitVarStrings.append(self.InitStrings[i])

            # In self.NetworkDict, clean rate equation parameter list of variables that occur in that reaction
            # Add FixedReagent to Params even if not a parameter in rate eqn (requirement to add '$' below)
            for id in list(self.NetworkDict.keys()):
                for reag in self.VarReagents:
                    if reag in self.NetworkDict[id]['Params']:
                        self.NetworkDict[id]['Params'].remove(reag)
                for reag in self.FixedReagents:
                    if (
                        reag + '[t]' in list(self.NetworkDict[id]['Reagents'].keys())
                    ) and (reag not in self.NetworkDict[id]['Params']):
                        self.NetworkDict[id]['Params'].append(reag + '[t]')

            # Warn if no reagents have been fixed
            if self.FixedReagents == []:
                print('Warning: No reagents have been fixed')
            else:  # Warn if a fixed reagent does not occur in a reaction equation
                for reag in self.FixedReagents:
                    if reag not in self.Reagents:
                        print(
                            'Warning: '
                            + reag
                            + ' (fixed) does not occur in any reaction'
                        )

            # Check whether all parameters have been initialised
            # johann -- remove [t] from params
            for id in list(self.NetworkDict.keys()):
                for i in range(len(self.NetworkDict[id]['Params'])):
                    self.NetworkDict[id]['Params'][i] = self.NetworkDict[id]['Params'][
                        i
                    ][:-3]
                    if self.NetworkDict[id]['Params'][i] not in self.InitParams:
                        print(
                            'Warning: Parameter '
                            + self.NetworkDict[id]['Params'][i]
                            + ' has not been initialised'
                        )

            # Check whether all variable reagents have been initialised
            for reag in self.VarReagents:
                if reag[:-3] + '[0]' not in self.Inits:
                    print('Warning: Variable ' + reag + ' has not been initialised')

            # Check that all initialised parameters actually occur in self.Inits
            known = 0
            for param in self.InitParams:
                for id in list(self.NetworkDict.keys()):
                    if param in self.NetworkDict[id]['Params']:
                        known = 1
                        break
                    else:
                        known = 0
                if not known:
                    print(
                        'Warning: '
                        + param
                        + ' has been initialised but does not occur in any rate equation'
                    )

            # clean up rate equations in self.NetworkDict to remove [t] for Params
            # clean up Reagents to remove [t] and add $ for fixed
            for id in list(self.NetworkDict.keys()):
                for param in self.NetworkDict[id]['Params']:
                    self.NetworkDict[id]['RateEq'] = self.NetworkDict[id][
                        'RateEq'
                    ].replace(param + '[t]', param)
                for reag in list(self.NetworkDict[id]['Reagents'].keys()):
                    if reag[:-3] in self.NetworkDict[id]['Params']:
                        saveval = self.NetworkDict[id]['Reagents'].pop(reag)
                        self.NetworkDict[id]['Reagents']['$' + reag[:-3]] = saveval
                    else:
                        saveval = self.NetworkDict[id]['Reagents'].pop(reag)
                        self.NetworkDict[id]['Reagents'][reag[:-3]] = saveval

            # output errors
            if self.ParseErrors != []:
                print('Parse errors occurred: ', self.ParseErrors)

            # debugging
            if debug:
                print('\n\n\n')
                print('\nself.ReactionIDs: ', self.ReactionIDs)
                print('\nself.NetworkDict: ', self.NetworkDict)
                print('\nself.Names: ', self.Names)
                print('\nself.Inits: ', self.Inits)
                print('\nself.InitStrings: ', self.InitStrings)
                print('\nself.InitParStrings: ', self.InitParStrings)
                print('\nself.InitVarStrings: ', self.InitVarStrings)
                print('\nself.InitParams: ', self.InitParams)
                print('\nself.Reagents: ', self.Reagents)
                print('\nself.FixedReagents: ', self.FixedReagents)
                print('\nself.VarReagents: ', self.VarReagents)
                print('\nParseErrors: ', self.ParseErrors)

            # now write the jws output file
            filename = File[:-4]
            filename = self.chkjws(filename)
            go = 0
            loop = 0
            filex = ''
            while loop == 0:
                try:
                    filex = os.path.join(outdir, filename)
                    f = open(filex, 'r')
                    f.close()
                    inp = input('\nFile "' + filex + '" exists.\nOverwrite? ([y]/n) ')
                    if inp == 'y' or inp == '':
                        go = 1
                        loop = 1
                    elif inp == 'n':
                        filename = input(
                            '\nFile "' + filename + '" exists. Enter a new filename: '
                        )
                        go = 1
                        filex = os.path.join(outdir, filename)
                        filename = self.chkjws(filename)
                    else:
                        print('\nInvalid input')
                except:
                    print('\nFile "' + filex + '" does not exist, proceeding...')
                    loop = 1
                    go = 1
            if go == 1:
                try:
                    UseR = getuser()
                except:
                    UseR = ''

                outFile = open(filex, 'w')
                header = ''
                # header +=  '############################################################\n'
                header += '# JWS model input file \n'
                header += (
                    '# Generated by PySCeS ('
                    + __version__
                    + ') (http://pysces.sourceforge.net) \n'
                )
                header += '# Pysces input file: ' + File + '\n'
                header += (
                    '# This file generated: '
                    + strftime("%a, %d %b %Y %H:%M:%S")
                    + ' by '
                    + UseR
                    + ' \n'
                )
                header += (
                    '###########################################################\n\n'
                )
                outFile.write(header)

                # modelname
                modelname = File[:-4]
                outFile.write('begin name\n' + modelname + '\nend name\n\n')

                # reactions and rate equations
                reaction_list = []
                rateeq_list = []

                nd = self.NetworkDict
                reaclist = copy.copy(
                    list(nd.keys())
                )  # johann -- to sort self.ReactionIDs neatly ;-)
                reaclist.sort()
                for key in reaclist:  # key = reaction name
                    reagL = []
                    reagR = []
                    Req = copy.copy(nd[key]['RateEq'])
                    for reagent in nd[key]['Reagents']:
                        if nd[key]['Reagents'][reagent] > 0:
                            reagR.append(
                                '{'
                                + str(abs(nd[key]['Reagents'][reagent]))
                                + '}'
                                + reagent
                            )
                        elif nd[key]['Reagents'][reagent] < 0:
                            reagL.append(
                                '{'
                                + str(abs(nd[key]['Reagents'][reagent]))
                                + '}'
                                + reagent
                            )
                    substring = ''
                    count = 0
                    for x in reagL:
                        if count != 0:
                            substring += ' + '
                        substring += x.replace(' ', '')
                        count += 1
                    prodstring = ''
                    count = 0
                    for x in reagR:
                        if count != 0:
                            prodstring += ' + '
                        prodstring += x.replace(' ', '')
                        count += 1
                    symbol = ' = '
                    reaction_list.append(key + '\t' + substring + symbol + prodstring)
                    rateeq_list.append(key + ' = ' + Req)
                outFile.write('begin reactions\n')
                for x in reaction_list:
                    outFile.write(x + '\n')
                outFile.write('end reactions\n\n')
                outFile.write('begin rate equations\n')
                for x in rateeq_list:
                    outFile.write(x + '\n')
                outFile.write('end rate equations\n\n')

                # parameters
                outFile.write('begin parameters\n')
                for x in self.InitParStrings:
                    outFile.write(x + '\n')
                outFile.write('end parameters\n\n')

                # species initial values
                outFile.write('begin initial conditions\n')
                for x in self.InitVarStrings:
                    outFile.write(x + '\n')
                outFile.write('end initial conditions\n\n')

                # close output file
                outFile.close()

                # print to stdout if quiet is set to zero
                if quiet == 0:
                    print('\nModel name: ' + modelname)

                    print("\nReactions:")
                    for x in reaction_list:
                        print(x)

                    print("\nRate Equations:")
                    for x in rateeq_list:
                        print(x)

                    print('\nParameters:')
                    for x in self.InitParStrings:
                        print(x)

                    print('\nSpecies Initial Values:')
                    for x in self.InitVarStrings:
                        print(x)

    def chkjws(self, File):
        """
        chkjws(File)

        Checks if a filename has a .jws extension and adds one to the returned filename if needed

        Arguments:
        =========
        File: the filename to check

        """
        try:
            if File[-4:] == '.jws':
                pass
            else:
                print('Assuming extension is .jws')
                File += '.jws'
        except:
            print('Chkjws error')
        return File


if __name__ == '__main__':
    import os, sys
    from time import sleep

    inDiR = 'c://mypysces//pscmodels'
    outDiR = 'c://mypysces//jws'

    jwp = JWSParser()
    for mod in os.listdir(inDiR):
        jwp.psc2jws(mod, indir=inDiR, outdir=outDiR, quiet=1, debug=0)

    psp = PySCeSParser(debug=0)



"""
PySCeS - Python Simulator for Cellular Systems (http://pysces.sourceforge.net)

Copyright (C) 2004-2022 B.G. Olivier, J.M. Rohwer, J.-H.S Hofmeyr all rights reserved,

Brett G. Olivier (bgoli@users.sourceforge.net)
Triple-J Group for Molecular Cell Physiology
Stellenbosch University, South Africa.

Permission to use, modify, and distribute this software is given under the
terms of the PySceS (BSD style) license. See LICENSE.txt that came with
this distribution for specifics.

NO WARRANTY IS EXPRESSED OR IMPLIED.  USE AT YOUR OWN RISK.
Brett G. Olivier
"""
from __future__ import division, print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from pysces.version import __version__

__doc__ = '''PySCeS plotting module'''

import numpy, scipy
import os
import itertools
from time import strftime
from getpass import getuser


class PyscesGPlot:
    '''Prototype plotting class which wraps gnuplot functions and adds some stuff'''

    __version__ = __version__
    save_html_header = 1
    save_html_footer = 1
    mode_gnuplot4 = 1

    def __init__(self):
        import scipy.sandbox.gplt as gnuplot

        self.gnuplot = gnuplot

    def plot2D(
        self, ginput, x, ylist, cmdout=0, name=None, fmt='w l', ykey=None, log=0
    ):
        """
        plot2D(ginput,x,ylist,cmdout=0,name=None,fmt='w l',ykey=None,log=0)

        Plot a 2D graph using gnuplot

        Arguments:
        =========
        ginput: an array containing data to be plotted
        x: the index of the x-axis
        ylist: a list of y-axis indices
        cmdout [default=0]: do not plot but instead output the equivalent gnuplot command
        name [default=None]: the array name to be used with cmdout=1 (otherwise ginput)
        fmt [default='w l']: gnuplot line format string 'w l'=lines, 'w p'=points
        ykey [default=None]: list of y-axis key names (otherwise line index is used)
        log [default=0]: set x and y axis to linear (0) or log (1) scales

        """
        assert ginput.shape[0] >= 1, 'ginputRow must be >= 1'
        assert ginput.shape[1] > 1, 'ginputCol must be > 1'
        assert x <= ginput.shape[1], 'X index should be < ginputCol index'
        assert len(ylist) >= 1, 'Plotlist must contain a value'

        if ykey != None and len(ylist) != len(ykey):
            ykey = None

        pltcmd = 'self.gnuplot.plot('
        for dep in range(len(ylist)):
            pltcmd += (
                'ginput[:,' + str(int(x)) + '],ginput[:,' + str(int(ylist[dep])) + '],'
            )
            if ykey != None:
                tstr = str(ykey[dep])
            else:
                tstr = str(ylist[dep])
            pltcmd += '\'t \"' + tstr + '\" ' + fmt + '\','
        pltcmd += ')'
        # print pltcmd
        if cmdout and name != None:
            return pltcmd.replace('ginput', str(name))
        elif cmdout:
            return pltcmd
        else:
            eval(pltcmd)
            if log:
                self.gnuplot.logx('on')
                self.gnuplot.logy('on')
            else:
                self.gnuplot.logx('off')
                self.gnuplot.logy('off')

    def plot3D(
        self,
        ginput,
        x,
        y,
        z,
        cmdout=0,
        name=None,
        fmt='w l',
        zkey=None,
        arrayout=0,
        log=0,
    ):
        """
        plot3D(ginput,x,y,z,cmdout=0,name=None,fmt='w l',zkey=None,arrayout=0,log=0)

        Plot a 3D surface with gnuplot

        Arguments:
        =========
        ginput: an array containing data to be plotted
        x: x-axis index
        y: y-axis index
        z: z-axis index
        cmdout [default=0]: do not plot but instead output the equivalent gnuplot command
        name [default=None]: the array name to be used with cmdout=1 (otherwise ginput)
        fmt [default='w l']: gnuplot line format string 'w l'=lines, 'w p'=points
        zkey [default=None]: list of z-axis key names (otherwise line index is used)
        arrayout [default=0]: output the 3 column array of data used by scipy.plot3d
        log [default=0]: set x, y, z axis to linear (0) or log (1) scales

        """
        assert ginput.shape[0] >= 1, 'ginputRow must be >= 1'
        assert ginput.shape[1] > 1, 'ginputCol must be > 1'
        assert x <= ginput.shape[1], 'X index should be < ginputCol index'
        assert y <= ginput.shape[1], 'Y index should be < ginputCol index'
        assert z <= ginput.shape[1], 'Z index should be < ginputCol index'

        if zkey != None:
            tstr = str(zkey)
        else:
            tstr = str(z)

        pltcmd = 'self.gnuplot.plot3d('
        pltcmd += 'ginput[:,:3],'
        pltcmd += '\'t \"' + tstr + '\" ' + fmt + '\','
        pltcmd += ')'

        plotfile = []
        plotfile.append(ginput[:, x])
        plotfile.append(ginput[:, y])
        plotfile.append(ginput[:, z])
        ginput = scipy.transpose(scipy.array(plotfile))

        if cmdout and name != None:
            if arrayout:
                return pltcmd.replace('ginput', str(name)), ginput
            else:
                return pltcmd.replace('ginput', str(name))
        elif cmdout:
            if arrayout:
                return pltcmd, ginput
            else:
                return pltcmd
        else:
            eval(pltcmd)
            if log:
                self.gnuplot.logx('on')
                self.gnuplot.logy('on')
            else:
                self.gnuplot.logx('off')
                self.gnuplot.logy('off')

    def plotX(self, ginput, cmdout=0, name=None, fmt='w l', log=0):
        """
        plotX(ginput,cmdout=0,name=None,fmt='w l',log=0)

        Quick and dirty 2D plot where all other columns are plotted against the first column

        Arguments:
        =========
        ginput: an array containing data to be plotted
        cmdout [default=0]: do not plot but instead output the equivalent gnuplot command
        name [default=None]: the array name to be used with cmdout=1 (otherwise ginput)
        fmt [default='w l']: gnuplot line format string 'w l'=lines, 'w p'=points
        log [default=0]: set x and y axis to linear (0) or log (1) scales

        """
        assert (
            ginput.shape[0] >= 1 and ginput.shape[1] > 1
        ), 'array must have at least dim(1,2)'

        row, col = ginput.shape
        pltcmd = 'self.gnuplot.plot('
        for x in range(1, col):
            pltcmd += 'ginput[:,0],ginput[:,' + str(x) + '],'
            pltcmd += '\'t \"' + str(x) + '\" ' + fmt + '\''
            if x < col - 1:
                pltcmd += ','
        pltcmd += ')'

        if cmdout and name != None:
            return pltcmd.replace('ginput', str(name))
        elif cmdout:
            return pltcmd
        else:
            eval(pltcmd)
            if log:
                self.gnuplot.logx('on')
                self.gnuplot.logy('on')
            else:
                self.gnuplot.logx('off')
                self.gnuplot.logy('off')

    def __save_command__(self, FileName):
        """
        __save_command__(FileName)

        Selects the correct gnuplot command depending on gnuplot version set with mode_gnuplot4

        Arguments:
        =========
        FileName: the name of the resulting output fil

        """
        if self.mode_gnuplot4:
            self.gnuplot.output(
                FileName,
                'png',
                options='medium size 640,480 xffffff x000000 x404040\
                              xff0000 xffa500 x66cdaa xcdb5cd\
                              xadd8e6 x0000ff xdda0dd x9500d3',
            )
        else:
            self.gnuplot.save(FileName)

    def save(self, filename, path=None):
        """
        save(filename,path=None)

        Save the current active plot to a file

        Arguments:
        =========
        filename: the name of the output file
        path [default=None]: the directory where the file should be saved

        """
        try:
            if filename[-4:] != '.png':
                filename += '.png'
                print('File saved as: ' + filename)
        except:
            pass

        if path != None:
            filename = os.path.join(path, filename)
        else:
            filename = os.path.join(os.getcwd(), filename)
        self.__save_command__(filename)

    def save_html(self, imagename, File=None, path=None, name=None, close_file=1):
        """
        save_html(imagename,File=None,path=None,name=None,close_file=1)

        Save the current plot to a file plus an HTML file that includes the resulting image

        Arguments:
        =========
        imagename: the output image name
        File [default=None]: the HTML filename defaults to imagename
        path [default=None]: the output directory
        name [default=None]: the HTML title of the image
        close_file [default=1]: close the HTML (1) or leave it open for further editing (0)

        """
        imagename = str(imagename)
        if name != None:
            name = str(name)
        else:
            name = ''
        try:
            if imagename[-4:] != '.png':
                imagename += '.png'
                print('File saved as: ' + imagename)
        except:
            pass

        if path != None:
            imagenameout = os.path.join(path, imagename)
            if File == None:
                File = open(os.path.join(path, imagename[:-3] + 'html'), 'w')
        else:
            imagenameout = os.path.join(os.getcwd(), imagename)
            if File == None:
                File = open(os.path.join(os.getcwd(), imagename[:-3] + 'html'), 'w')
        self.__save_command__(imagename)

        fname = (
            'PySCeS generated image - '
            + imagename
            + '" generated from model file: '
            + strftime("%H:%M:%S")
        )

        if File != None:
            assert (
                File != file
            ), 'WriteArray(input,File=None,Row=None,Col=None,close_file=0)'
            header = '\n'
            header += (
                '<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN">\n'
            )
            header += '<html>\n'
            header += '<head>\n'
            header += (
                '<title>PySCeS generated image - '
                + imagename
                + ' - '
                + strftime("%H:%M:%S (%Z)")
                + '</title>\n'
            )
            header += '<meta http-equiv="Content-Type" content="text/html; charset=iso-8859-1">\n'
            header += '</head>\n'
            header += '<body bgcolor="#FFFFCC">\n\n'
            header += (
                '<h4><a href="http://pysces.sourceforge.net">PySCeS</a> generated image - '
                + imagename
                + '</h4>\n\n'
            )

            if self.save_html_header:
                File.write(header)
            File.write('\n<!-- ' + imagename + '-->\n\n')

            File.write('<p align="center">\n')

            File.write(
                '<img src="' + imagename + '" width="640" height="480" border="2">\n'
            )

            if name != None:
                File.write('<br><font size="3">' + str(name) + '</font>\n')

            File.write('</p>\n\n')
            if self.save_html_footer:
                try:
                    File.write(
                        '<p><a href="http://pysces.sourceforge.net"><font size="3">PySCeS '
                        + self.__version__
                        + '</font></a><font size="2"> HTML output (image <i>'
                        + imagename
                        + '</i> produced at '
                        + strftime("%H:%M:%S")
                        + ' by <i>'
                        + getuser()
                        + '</i>)</font></p>\n'
                    )
                except:
                    File.write(
                        '<p><a href="http://pysces.sourceforge.net"><font size="3">PySCeS '
                        + __version__
                        + '</font></a><font size="2"> HTML output (model <i>'
                        + imagename
                        + '</i> produced at '
                        + strftime("%H:%M:%S - %Z")
                        + ')</font></p>\n'
                    )
                File.write('</body>\n')
                File.write('</html>\n')
            else:
                File.write('\n')

            if close_file:
                File.close()

    def gridon(self):
        """
        gridon()

        Enable grid for the current active plot

        Arguments:
        None

        """
        self.gnuplot.grid('on')

    def gridoff(self):
        """
        gridoff()

        Disable grid for the current active plot

        Arguments:
        None

        """
        self.gnuplot.grid('off')

    def logx(self):
        """
        logx()

        Set x logscale for active plot

        Arguments:
        None

        """
        self.gnuplot.logx('on')

    def logy(self):
        """
        logy()

        Set y logscale for active plot

        Arguments:
        None

        """
        self.gnuplot.logy('on')

    def logxy(self):
        """
        logxy()

        Set x and y logscale for active plot

        Arguments:
        None

        """
        self.gnuplot.logx('on')
        self.gnuplot.logy('on')

    def linx(self):
        """
        linx()

        Set x linear scale for active plot

        Arguments:
        None

        """
        self.gnuplot.logx('off')

    def liny(self):
        """
        liny()

        Set y linear scale for active plot

        Arguments:
        None

        """
        self.gnuplot.logy('off')

    def linxy(self):
        """
        linxy()

        Set x and y linear scale for active plot

        Arguments:
        None

        """
        self.gnuplot.logx('off')
        self.gnuplot.logy('off')

    def logxliny(self):
        """
        logxliny()

        Set logscale x and linear scale y for currently active plot

        Arguments:
        None

        """
        self.gnuplot.logx('on')
        self.gnuplot.logy('off')

    def linxlogy(self):
        """
        linxlogy()

        Set logscale y and linear scale x for currently active plot

        Arguments:
        None

        """
        self.gnuplot.logx('off')
        self.gnuplot.logy('on')

    def xrng(self, start, end):
        """
        xrng(start,end)

        Set the range of the x-axis for currently active plot

        Arguments:
        =========
        start: lower value
        end: upper value

        """
        self.gnuplot.xaxis((float(start), float(end)))

    def yrng(self, start, end):
        """
        yrng(start,end)

        Set the range of the y-axis for currently active plot

        Arguments:
        =========
        start: lower value
        end: upper value

        """
        self.gnuplot.yaxis((float(start), float(end)))

    def zrng(self, start, end):
        """
        zrng(start,end)

        Set the range of the z-axis for currently active plot

        Arguments:
        =========
        start: lower value
        end: upper value

        """
        self.gnuplot.zaxis((float(start), float(end)))

    def xlabel(self, l=''):
        """
        xlabel(l='')

        Set the x-axis label for the currently active plot

        Arguments:
        =========
        l [default=' ']: axis label string

        """
        self.gnuplot.xtitle(str(l))

    def ylabel(self, l=''):
        """
        ylabel(l='')

        Set the y-axis label for the currently active plot

        Arguments:
        =========
        l [default=' ']: axis label string

        """
        self.gnuplot.ytitle(str(l))

    def zlabel(self, l=''):
        """
        zlabel(l='')

        Set the z-axis label for the currently active plot

        Arguments:
        =========
        l [default=' ']: axis label string

        """
        self.gnuplot.ztitle(str(l))

    def title(self, l=''):
        """
        title(l='')

        Set the graph title for the currently active plot

        Arguments:
        =========
        l [default=' ']: graph label string

        """
        self.gnuplot.title(str(l))

    def ticslevel(self, x=0.0):
        """
        ticslevel(x=0.0)

        Set the gnuplot ticlevel for the currently active plot

        Arguments:
        =========
        x [default=0.0]: gnuplot ticlevel

        """
        self.gnuplot.ticlevel(float(x))


class LineObj(object):
    __id__ = None
    idx = None
    data = None
    prop = None

    def __init__(self, data, idx, **kwargs):
        self.idx = idx
        self.data = data
        self.prop = {
            'label': None,
            'linewidth': None,
            'colour': None,
            'style': None,
        }
        for k in list(kwargs.keys()):
            if k in self.prop:
                self.prop[k] = kwargs[k]


class AxisObj(object):
    __id__ = None
    idx = None
    data = None
    prop = None

    def __init__(self, data, idx, **kwargs):
        self.idx = idx
        self.data = data
        self.prop = {
            'label': None,
            'log': False,
        }
        for k in list(kwargs.keys()):
            if k in self.prop:
                self.prop[k] = kwargs[k]


class GraphicsObj(object):
    xaxis = None
    yaxis = None
    data = None
    data_shape = None
    data_type = None
    data_lines = None
    var_idx = None
    scalars = None

    def setData(self, data):
        assert type(data) == numpy.ndarray, '\nI need a numpy/scipy ndarray!'
        assert data.shape[1] > 1, '\nI need at least a 2D array!'
        self.data = data
        self.data_shape = data.shape
        self.data_type = data.dtype.char
        self.scalars = numpy.zeros((data.shape[0], 1), 'd')
        self.resetAxis()
        self.setDataLabels(['l' + str(l) for l in range(self.data_shape[1])])

    def setDataLabels(self, labels):
        assert len(labels) == self.data_shape[1], '\nList unacceptable length.'
        if self.data_lines != None:
            [self.__delattr__(n) for n in self.data_lines]
        self.data_lines = []
        for l in range(len(labels)):
            self.data_lines.append(labels[l])
            ##  lo = LineObj(self.data[:,l].reshape(self.data_shape[0], 1), l, label=labels[l])
            lo = LineObj(numpy.take(self.data, [l], axis=1), l, label=labels[l])
            setattr(self, labels[l], lo)

    def setAxis(self, **kwargs):
        ##  print kwargs
        for key in list(kwargs.keys()):
            if key in ['x', 'y'] and kwargs[key] in self.var_idx:
                if getattr(self, key + 'axis') != None:
                    self.var_idx.append(getattr(self, key + 'axis').idx)
                    self.var_idx.sort()
                    ##  print '\t', self.var_idx
                ##  ao = AxisObj(self.data[:,kwargs[key]].reshape(self.data_shape[0], 1), kwargs[key], label=key)
                ao = AxisObj(
                    numpy.take(self.data, [kwargs[key]], axis=1), kwargs[key], label=key
                )
                setattr(self, key + 'axis', ao)
                a = self.var_idx.pop(self.var_idx.index(kwargs[key]))
                ##  print '\tkey idx :', key, kwargs[key]
        ##  print 'var_idx', self.var_idx

    def resetAxis(self):
        self.xaxis = self.yaxis = None
        self.var_idx = list(range(self.data_shape[1]))

    def getIdx(self, label):
        assert (
            label in self.data_lines
        ), '\n%s is not a valid data line.\nSelect one of %s' % (label, self.data_lines)
        return self.data_lines.index(label)

    def setData2Scalar(self, idx):
        self.scalars = self.data.take([idx], axis=1)

    def dataMinMax(self, idx):
        return (numpy.min(self.data[:, idx]), numpy.max(self.data[:, idx]))

    def dataLog10(self, idx):
        self.data[:, idx] = numpy.log10(self.data[:, idx])

    def getVarData(self):
        return self.data.take(self.var_idx, axis=1)

    def getRawData(self):
        return {
            'data': self.data,
            'scalars': self.scalars,
            'x_idx': self.xaxis.idx,
            'y_idx': self.yaxis.idx,
            'var_idx': self.var_idx,
            'data_lines': self.data_lines,
        }


class Graphics2dObj(GraphicsObj):
    title = 'Graph'
    xtitle = 'x'
    ytitle = 'y'
    xlog = False
    ylog = False
    selected = None
    __dims__ = 2

    def __init__(self, data, xidx=None, labels=None, selected=None):
        self.setData(data)
        if xidx != None:
            self.setAxis(x=xidx)
        if labels != None:
            self.setDataLabels(labels)
        if selected != None:
            self.selected = selected
        else:
            selected = []

    # working on this ... think I'm almost finished now
    def getNewByIndex(self, x, lines):
        data = self.data.take([x] + lines, axis=1)
        G = Graphics2dObj(data, 0)
        G.setDataLabels([self.data_lines[x]] + [self.data_lines[i] for i in lines])
        return G

    # working on this ... think I'm almost finished now
    def getNewByName(self, x, lines):
        return self.getNewByIndex(self.getIdx(x), [self.getIdx(l) for l in lines])


class Graphics3dObj(Graphics2dObj):
    ztitle = 'z'
    zlog = False
    __dims__ = 3

    def __init__(self, data, xidx, yidx):
        self.setData(data)
        self.setAxis(x=xidx, y=yidx)


class PyscesGPlot2MPL:
    """
    Old 'gnuplot' plotting functions using a matplotlib backend
    """

    mode_interactive = True
    mode_create_new_figure = True
    new_figure_generator = None
    current_figure = None
    mode_hold = False
    # backwards compatibility
    save_html_header = 1
    save_html_footer = 1
    max_open_windows = 50

    def __init__(self):
        import matplotlib

        try:
            matplotlib.use('TKagg')
        except Exception as ex:
            print(ex)
            print(
                "\nPySCeS uses matplotlib's TKagg backend for interactive plotting please enable this backend when compiling matplotlib"
            )
        ## import matplotlib.axes3d
        import pylab

        ## self._matplotlib_axes3d_ = matplotlib.axes3d
        self.P = pylab
        if self.mode_interactive:
            self.P.ion()
        self.setNewFigureGenerator(self.max_open_windows)

    def setNewFigureGenerator(self, num):
        """
        Create a new_figure_generator with range [1,num+1]
        """
        self.new_figure_generator = itertools.cycle(list(range(1, num + 1)))
        self.max_open_windows = num

    def setActiveFigure(self, fignum):
        self.current_figure = self.P.figure(fignum)

    def setNewFigure(self):
        self.setActiveFigure(next(self.new_figure_generator))

    def closeAllPlotWindows(self):
        for x in range(self.max_open_windows):
            self.P.close()

    def plot2D(self, data, x, ylist=[], labels=None, style=None):
        assert type(ylist) == list, '\nylist must be a list'
        if self.mode_create_new_figure:
            self.setNewFigure()
        if not self.mode_hold:
            self.P.clf()
        D = Graphics2dObj(data, x)
        if labels != None:
            D.setDataLabels(labels)
        if len(ylist) == 0:
            print('ylist empty plotting all data vs (%s)' % x)
            ylist = list(range(data.shape[1]))
            ylist.pop(ylist.index(x))
        if len(ylist) > 0:
            ylt = numpy.array(ylist) < D.data_shape[1]
            assert (
                type(ylt) == numpy.ndarray
            ), '\nInvalid y list this can be caused by a forgotten keyword labels='
            assert ylt.all(), '\nInvalid y index'
            D = D.getNewByIndex(x, ylist)
        self.P.ioff()
        usedline = []
        for line in D.data_lines:
            L = getattr(D, line)
            ##  print line, D.xaxis.idx, L.idx
            if L.idx != D.xaxis.idx and line not in usedline:
                ##  print 'done'
                usedline.append(line)
                if style != None:
                    self.P.plot(D.xaxis.data, L.data, style, label=L.prop['label'])
                else:
                    self.P.plot(D.xaxis.data, L.data, label=L.prop['label'])
        if self.mode_interactive:
            self.P.ion()
        self.P.legend()

    def plot3D(self, *args, **kwargs):
        print("*****\nplot3D not implemented yet.\n*****")

    def plotX(self, data):
        if self.mode_create_new_figure:
            self.setNewFigure()
        if not self.mode_hold:
            self.P.clf()
        D = Graphics2dObj(data, None)
        ##  print type(D)
        ##  print D.var_idx
        ##  print D.data[1]
        self.P.plot(D.getVarData())

    def save(self, fname, path=None):
        fname += '.png'
        if path != None:
            fname = os.path.join(path, fname)
        self.P.savefig(fname)

    def legendOff(self):
        L = self.P.gca().get_legend()
        L._visible = False
        self.P.draw()

    def gridon(self):
        self.P.grid(True)

    def gridoff(self):
        self.P.grid(False)

    def logx(self):
        A = self.P.gca()
        A.set_xscale('log')
        self.P.draw()

    def logy(self):
        A = self.P.gca()
        A.set_yscale('log')
        if self.mode_interactive:
            self.P.draw()

    def linx(self):
        A = self.P.gca()
        A.set_xscale('linear')
        if self.mode_interactive:
            self.P.draw()

    def liny(self):
        A = self.P.gca()
        A.set_yscale('linear')
        if self.mode_interactive:
            self.P.draw()

    def linxy(self):
        A = self.P.gca()
        A.set_xscale('linear')
        A.set_yscale('linear')
        if self.mode_interactive:
            self.P.draw()

    def logxy(self):
        A = self.P.gca()
        A.set_xscale('log')
        A.set_yscale('log')
        if self.mode_interactive:
            self.P.draw()

    def logxliny(self):
        A = self.P.gca()
        A.set_xscale('log')
        A.set_yscale('linear')
        if self.mode_interactive:
            self.P.draw()

    def linxlogy(self):
        A = self.P.gca()
        A.set_xscale('linear')
        A.set_yscale('log')
        if self.mode_interactive:
            self.P.draw()

    def xrng(self, start, end):
        self.P.xlim(start, end)

    def yrng(self, start, end):
        self.P.ylim(start, end)

    def zrng(self, start, end):
        print("*****\n\tNot implemented yet.\n*****")

    def xlabel(self, l=''):
        self.P.xlabel(l)

    def ylabel(self, l=''):
        self.P.ylabel(l)

    def zlabel(self, l=''):
        print("*****\n\tNot implemented yet.\n*****")

    def title(self, l=''):
        self.P.title(l)

    def ticslevel(self, x=0.0):
        print("*****\n\tNot implemented yet.\n*****")

    def setLineWidth(self, w=1):
        'set line width for current axis'
        A = self.P.gca()
        [l.set_linewidth(w) for l in A.get_lines()]
        if self.mode_interactive:
            self.P.draw()

    def save_html(self, imagename, File=None, path=None, name=None, close_file=1):
        imagename = str(imagename)
        if name != None:
            name = str(name)
        else:
            name = ''
        try:
            if imagename[-4:] != '.png':
                imagename += '.png'
                print('File saved as: ' + imagename)
        except:
            pass

        if path != None:
            imagenameout = os.path.join(path, imagename)
            if File == None:
                File = open(os.path.join(path, imagename[:-3] + 'html'), 'w')
        else:
            imagenameout = os.path.join(os.getcwd(), imagename)
            if File == None:
                File = open(os.path.join(os.getcwd(), imagename[:-3] + 'html'), 'w')
        self.P.savefig(imagenameout, dpi=80)

        fname = (
            'PySCeS generated image - '
            + imagename
            + '" generated from model file: '
            + strftime("%H:%M:%S")
        )

        if File != None:
            assert (
                File != file
            ), 'WriteArray(input,File=None,Row=None,Col=None,close_file=0)'
            header = '\n'
            header += (
                '<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN">\n'
            )
            header += '<html>\n'
            header += '<head>\n'
            header += (
                '<title>PySCeS generated image - '
                + imagename
                + ' - '
                + strftime("%H:%M:%S (%Z)")
                + '</title>\n'
            )
            header += '<meta http-equiv="Content-Type" content="text/html; charset=iso-8859-1">\n'
            header += '</head>\n'
            header += '<body bgcolor="#FFFFCC">\n\n'
            header += (
                '<h4><a href="http://pysces.sourceforge.net">PySCeS</a> generated image - '
                + imagename
                + '</h4>\n\n'
            )

            if self.save_html_header:
                File.write(header)
            File.write('\n<!-- ' + imagename + '-->\n\n')

            File.write('<p align="center">\n')

            File.write(
                '<img src="' + imagename + '" width="640" height="480" border="2">\n'
            )

            if name != None:
                File.write('<br><font size="3">' + str(name) + '</font>\n')

            File.write('</p>\n\n')
            if self.save_html_footer:
                try:
                    File.write(
                        '<p><a href="http://pysces.sourceforge.net"><font size="3">PySCeS '
                        + self.__version__
                        + '</font></a><font size="2"> HTML output (image <i>'
                        + imagename
                        + '</i> produced at '
                        + strftime("%H:%M:%S")
                        + ' by <i>'
                        + getuser()
                        + '</i>)</font></p>\n'
                    )
                except:
                    File.write(
                        '<p><a href="http://pysces.sourceforge.net"><font size="3">PySCeS '
                        + __version__
                        + '</font></a><font size="2"> HTML output (model <i>'
                        + imagename
                        + '</i> produced at '
                        + strftime("%H:%M:%S - %Z")
                        + ')</font></p>\n'
                    )
                File.write('</body>\n')
                File.write('</html>\n')
            else:
                File.write('\n')

            if close_file:
                File.close()


class PyscesGPlot2MplExt(PyscesGPlot2MPL):
    def __init__(self):
        PyscesGPlot2MPL.__init__(self)

    def plotLines2DX(self, data, x, labels=None):
        if not self.mode_hold:
            self.P.clf()
        ##  F = self.P.figure(1)
        ##  S = self.P.subplot(1,1,1)
        D = Graphics2dObj(data, x)
        print(D.var_idx)
        print(D.data[1])
        self.P.plot(D.xaxis.data, D.getVarData())



import os
##  os.chdir('c:/mypysces/kraken')
os.chdir('/home/bgoli/mypysces/kraken')
from Kraken import *
import itertools
import scipy.io

#investigate using vstack with type/range checking
def concatenateArrays(array_list):
    output = None
    for arr in range(len(array_list)):
        if type(array_list[arr]) == numpy.ndarray:
            if arr == 0:
                output = array_list[arr]
            else:
                output = scipy.concatenate((output,array_list[arr]))
    return output

def splitScan(start, end, intervals, job, log=False):
    """Splits a range into a number of jobs with intervals"""
    assert intervals >= 1, '\n* Minimum of 1 interval'
    if log:
        kpoints = scipy.logspace(scipy.log10(start), scipy.log10(end), intervals+1)
    else:
        kpoints = scipy.linspace(start, end, intervals+1)
    job_list = []
    for p in range(len(kpoints)-1):
        job2 = job % (kpoints[p], kpoints[p+1])
        job_list.append(job2)
        print(job2)
    return job_list

def buildCycler(lst):
    """
    creates an infinite looping generator with itertools
    """
    return itertools.cycle(lst)

def deltaCommand():
    """
    This is used to setup conditions for each next-job-on-server
    """
    delta = 'P_NONE'
    return [delta]


def Scheduler1(contrl, job_list, init_list, task_id):
    """
    contrl is a getBasicController() instance
    job_list is a list of jobs to run
    init_list set of `static' command used to initialise each job
    """
    original_job_list = tuple(job_list)
    TIME_START = time.time()
    DaResults = []
    JobsWaiting = True
    getServer = buildCycler(contrl.available_server_list)
    while JobsWaiting:
        deferred_jobs = []
        job_servers = []
        for job in range(len(job_list)):
            if job < len(contrl.available_server_list):
                print('\nJob %s queued for processing...' % (job+1))
                server = next(getServer)
                job_servers.append(server)
                print("\nProcessing job %s init list ..." % (job+1))
                contrl.sendCmdListToOne(init_list, server)
                print("\nProcessing job %s delta command ..." % (job+1))
                contrl.sendCmdListToOne(deltaCommand(), server)
                contrl.sendJobToOne(job_list[job], server)
            else:
                deferred_jobs.append(job_list[job])
                print('Job %s deferred and rescheduled.' % (job+1))
        print("\nProcessing queued jobs ...")
        contrl.runAllJobs()
        for server in job_servers:
            contrl.getDataFromOneJob(server)
            arr = contrl.tentacle_response[server]
            DaResults.append(arr)
            print(type(arr))
        contrl.clearProcessMap()
        if len(deferred_jobs) > 0:
            print('\nDeferred:', deferred_jobs)
            job_list = deferred_jobs
        else:
            JobsWaiting = False
        cPickle.dump(DaResults, open(task_id+'_result_list.bin','wb'), 2)

    TIME_END = time.time()

    print('\nNumber of results = %s' % len(DaResults))
    total_states = 0
    for x in range(len(DaResults)):
        try:
            print('\tResult %s has %s rows and %s columns' % (x+1, DaResults[x][0].shape[0], DaResults[x][0].shape[1]))
            total_states += DaResults[x][0].shape[0]
        except:
            print(type(DaResults[x][0]))
            print('\tResult %s %s: %s' % (x+1, type(DaResults[x][0]), DaResults[x][0]))
            print(original_job_list[x])


    print('Total time taken to complete %s state scan %s minutes' % (total_states, (TIME_END-TIME_START)/60.0))
    return DaResults

def GridSortLR(arr):
    """
    Sort irregular unstructured grid data into monotonically rising grid data
    (where possible) Work from left column to right (uses ultra cool NumPy functions :-)
    """
    print('arr.shape:', arr.shape)
    sorted = numpy.rot90(arr)
    print('rotated.shape:', sorted.shape)
    sorted_idx = numpy.lexsort(sorted)
    print('sorted_idx.shape', sorted_idx.shape)
    sorted = sorted.take(sorted_idx, axis=-1)
    sorted = numpy.flipud(sorted)
    arr = numpy.transpose(sorted)
    print('arr.shape:', arr.shape)
    return arr

def ScalarFunc(val):
    """
    Manipulate the scalar (or z-axis) values
    Accepts and returns a double
    """
    return val

def writeGnuPlot_3D(fname, arr):
    scipy.io.write_array(fname+'_result_gplt.dat', arr, separator=' ')

def writeVTK_UnstructuredGrid(fname, arr):
    """
    Write arr as a VTK unstructured grid to file "fname.vtk"
    arr must either be in:
        x,y,z format (in which case z is used for scalar data) or
        z,y,z,s format (s being point scalar value)
    Overwrite the ScalarFunc(val) function to manipulate the scalar (z-axis) values

    Original version http://mayavi.sourceforge.net/mwiki/PointClouds
    Hacked quite a bit by brett 20070221
    """
    assert arr.shape[1] == 3 or arr.shape[1] == 4, '\nneed s or for columns for this'
    if arr.shape[1] == 4:
        HAVE_SCALARS = 1
    else:
        HAVE_SCALARS = 0
        print('No scalar values supplied Z axis values will be used')

    n=arr.shape[0]
    print("n:",n)
    # write data to vtk polydata file
    # write header
    out = open(fname+'.vtk', 'w')
    h1 = "# vtk DataFile Version 2.0\n"
    h1 += "%s\n" % fname
    h1 += "ASCII\n"
    h1 += "DATASET UNSTRUCTURED_GRID\n"
    h1 += "POINTS " + str(n) + " double\n"
    out.write(h1)
    # write xyz data
    for r in range(n):
        #s = '%15.2f %15.2f %15.2f' % (x[i], y[i], z[i])
        out.write(str(arr[r,0])+" "+str(arr[r,1])+" "+str(arr[r,2])+'\n')

    # write cell data
    out.write("CELLS "+ str(n)+ " "+ str(2*n)+'\n')
    for r in range(n):
            #s = '1 %d \n' % (i)
            out.write("1 "+str(r)+"\n")

    # write cell types
    out.write("CELL_TYPES " + str(n)+'\n')
    for r in range(n):
        out.write("1 \n")

    # write z scalar values
    h2 = '\n' + """POINT_DATA """ + str(n) + "\n"
    h3 = "SCALARS %s double 1\n" % fname
    h3 += "LOOKUP_TABLE default\n"
    out.write(h2 + h3)

    for r in range(n):
        if HAVE_SCALARS:
            sc=(ScalarFunc(arr[r,3]))
        else:
            sc=(ScalarFunc(arr[r,2]))
        out.write(str(sc)+ "\n")

    out.write('\n')
    out.close()




# Parameter scan
if __name__ == '__main__':
    import time
    print(time.strftime('%y%m%d-%H%M'))
    START = time.time()

    """
    Build controller and initialise active tentacles
    """
    blob = getBasicController()
    Model_File = '2006_isola2.psc'
    ##  Model_Dir = 'c:\\mypysces\\pscmodels\\isola06'
    Model_Dir = '/home/bgoli/mypysces/pscmodels/isola06'
    blob.startModelServer(Model_File, Model_Dir)
    blob.startTentacleMonitor(interval=120)
    print('Server list: %s\n' % blob.available_server_list)

    """
    Get a name for this task (task == collection of jobs)
    """
    task_name = 'simple_scan'

    """
    Set up a list of 'static' initialisation commands that will be used to
    initialise each server before a job is sent to it
    """
    I_list = ['P_INIT,' + blob.model_server.model_file_name, 'P_LOAD',\
    'P_SET_FLOAT,mode_state_init,3', 'P_SET_FLOAT,nleq2_nitmax,20',\
    'P_SET_FLOAT,mode_state_init3_factor,0.6',\
    'P_SET_FLOAT,V2,200.0', 'P_SET_FLOAT,V3,200.0']

    pointsP1 = 50
    pointsP2 = 500
    scan_segments = 100
    scan_job = 'P_SCAN, 2, 9, P, %s, %s, '+str(pointsP1)+', 1, 0, A05, 0.01, 10000.0,  '+str(pointsP2)+', 1, 0,  A05, P, J_R1, V2, V3, lambda1, lambda2, eigen_sum, eigen_prod'
    J_list = splitScan(0.01, 3.7e4, scan_segments, scan_job, log=True)


    """
    Feed the controller, job list, initiation list and task name to the scheduler
    and off we go, we end up with a list of results
    """
    DaResults = Scheduler1(blob, J_list, I_list, task_name)
    DaResults = concatenateArrays(DaResults)
    print('\nDaResults.shape', DaResults.shape)
    print('DaResults[0]\n', DaResults[0])

    print(time.strftime('%y%m%d-%H%M'))
    MID = time.time()

    coords = numpy.transpose(numpy.array((DaResults[:,0],DaResults[:,1],DaResults[:,2]),'d'))
    print('\ncoords.shape', coords.shape, coords.dtype)
    print('coords[0]\n', coords[0])
    writeGnuPlot_3D(task_name, coords)
    for r in range(coords.shape[0]):
        coords[r,:] = scipy.log10(coords[r,:])
    writeVTK_UnstructuredGrid(task_name, coords)

    print(time.strftime('%y%m%d-%H%M'))
    END = time.time()

    print("\n********** start Kraken report **********")
    print("Using %s tentacles" % len(blob.available_server_list))
    print("Scan points          = %s" % (pointsP1*pointsP2*scan_segments))
    print("Data generation time = %2.2f minutes." % ((MID-START)/60.0))
    print("Data analysis time   = %2.2f minutes." % ((END-MID)/60.0))
    print("Total scan time      = %2.2f minutes." % ((END-START)/60.0))
    print("Total points/minute  = %2.1f" % ((60.0*pointsP1*pointsP2*scan_segments)/(END-START)))
    print("********** end Kraken report **********\n")



# Continuation
'''
if __name__ == '__main__':
    """
    Build controller and initialise active tentacles
    """
    blob = getBasicController()
    Model_File = '2006_isola2.psc'
    Model_Dir = 'c:\\mypysces\\pscmodels\\isola06'
    ##  Model_Dir = '/home/bgoli/mypysces/pscmodels/isola06'
    blob.startModelServer(Model_File, Model_Dir)
    blob.startTentacleMonitor(interval=60)
    print 'Server list: %s\n' % blob.available_server_list

    """
    Get a name for this task (task == collection of jobs)
    """
    task_name = 'cont_eigen'

    """
    Set up a list of 'static' initialisation commands that will be used to
    initialise each server before a job is sent to it
    """
    I_list = ['P_INIT,' + blob.model_server.model_file_name, 'P_LOAD',\
        'P_SET_FLOAT,mode_state_init,3', 'P_SET_FLOAT,nleq2_nitmax,20',\
        'P_SET_FLOAT,mode_state_init3_factor,0.5',\
        'P_SET_FLOAT,V2,200.0', 'P_SET_FLOAT,V3,200.0', 'P_SET_FLOAT,S05,1.0']
        ##  'P_SET_FLOAT,V2,1000.0', 'P_SET_FLOAT,V3,1000.0', 'P_SET_FLOAT,S05,1.0']

    """
    Set up our continuation:
    We want a parameter to be set with every new job/server and make use
    of the deltaCommand() function. Here we use it to set our y-axis (A05) values
    """
    yrange = 96
    xpoints = 64

    # create a list of values
    delta_list = scipy.logspace(scipy.log10(0.01),scipy.log10(10000.0), yrange)
    # initiate an 'looping' generator with it
    delta_gen = buildCycler(delta_list)

    def deltaCommand():
        """
        This is used to setup conditions for each next-job-on-server
        Used in sendCmdListToOne() so may one or more commands returned
        as a list
        """
        delta = 'P_SET_FLOAT,A05,%s' % delta_gen.next()
        # must return a list !!!
        return [delta]

    """
    Set up our list of jobs, in this case a single continuation+eigen with
    density xpoints at each of the delta (A05 values)
    """
    J_list = []
    for d in delta_list:
        ##  J_list.append('P_CONTINUATION,P,0.001,4.0e2,%s,%s' % (xpoints,d))
        J_list.append('P_CONTINUATION_WITH_EIGEN,P,0.001,3.0e4,%s,%s' % (xpoints,d))

    """
    Feed the controller, job list, initiation list and task name to the scheduler
    and off we go, we end up with a list of results
    """
    DaResults = Scheduler1(blob, J_list, I_list, task_name)

    """
    From here on we are processing/formatting our result data
    """
    iRes = []
    sRes = []
    jRes = []
    eRes = []
    for res in DaResults:
        iRes.append(res[0])
        sRes.append(res[1])
        jRes.append(res[2])
        eRes.append(res[3])
    del DaResults

    FinalIArray = concatenateArrays(iRes)
    FinalSArray = concatenateArrays(sRes)
    FinalJArray = concatenateArrays(jRes)
    FinalEArray = concatenateArrays(eRes)
    print FinalIArray.shape, FinalIArray.dtype
    print FinalSArray.shape, FinalSArray.dtype
    print FinalJArray.shape, FinalJArray.dtype
    print FinalEArray.shape, FinalEArray.dtype
    del iRes
    del sRes
    del jRes
    del eRes

    cPickle.dump((FinalIArray,FinalSArray,FinalJArray,FinalEArray), open(task_name+'_result_array.bin','wb'), 2)

    # get rid of negative values in axes and fluxes
    FinalArrayClean = []
    FinalVTK = []
    for r in range(FinalSArray.shape[0]):
        if FinalIArray[r,:].all() > 0.0 and FinalSArray[r,:].all() > 0.0 and FinalJArray[r,:].all() > 0.0:
            FinalArrayClean.append([FinalIArray[r,0], FinalIArray[r,1], FinalJArray[r,1], max(FinalEArray[r,:].real)])
        else:
            FinalIArray[r,:] == 0.0
            FinalSArray[r,:] == 0.0
            FinalJArray[r,:] == 0.0
            FinalEArray[r,:] == 0.0
    FinalArrayClean = numpy.array(FinalArrayClean)


    # write out an xyz gplt file
    coords = numpy.transpose(numpy.array((FinalArrayClean[:,0],FinalArrayClean[:,1],FinalArrayClean[:,2])))
    # get our data in nice grid format for plotting
    coords = GridSortLR(coords)
    writeGnuPlot_3D(task_name, coords)

    def ScalarFunc(val):
        """
        Manipulate the scalar (z-axis) values
        Accepts and returns a double
        """
        if val >= 0.0:
            return 100.0
        else:
            return val


    # prepare vtk data
    vtkcoords = numpy.transpose(numpy.array((FinalArrayClean[:,0],FinalArrayClean[:,1],FinalArrayClean[:,2],FinalArrayClean[:,3]),'d'))
    for row in range(vtkcoords.shape[0]):
        vtkcoords[row,:3] = scipy.log10(vtkcoords[row,:3])

    print vtkcoords[:10,:]
    vtkcoords = GridSortLR(vtkcoords)
    print vtkcoords[:10,:]

    writeVTK_UnstructuredGrid(task_name+'_result_ug', vtkcoords)
'''



#!/usr/bin/env python
# Testing the new parallel scanner class

import os

backupdir = os.getcwd()

import numpy as np
import pysces

tbox = pysces.PyscesUtils.TimerBox()
import time

m = pysces.model('isola2a')

ser = pysces.Scanner(m)

print("Serial execution...")
print("Start: ", tbox.normal_timer('SER'))
print(next(tbox.SER))
t1 = time.time()
ser.quietRun = True
ser.addScanParameter('V4', 60, 100, 11)
ser.addScanParameter('V1', 100, 160, 16)
ser.addScanParameter('V2', 100, 130, 16, follower=True)
ser.addScanParameter('V3', 80, 90, 6)
ser.addUserOutput('J_R1', 'A', 'ecR4_X', 'ccJR1_R1')
# ser.addUserOutput('J_R1', 'A')
ser.Run()
print("Done: ", next(tbox.SER))
t2 = time.time()
print("Duration: %.2f seconds" % (t2 - t1))
ser.statespersecond = len(ser.ScanSpace) / (t2 - t1)
print("States per second: %.1f" % ser.statespersecond)

print("\n\nParallel execution...scans per run =", 100)
par = pysces.ParScanner(m, engine='multiproc')
par.scans_per_run = 100
t3 = time.time()
par.addScanParameter('V4', 60, 100, 11)
par.addScanParameter('V1', 100, 160, 16)
par.addScanParameter('V2', 100, 130, 16, follower=True)
par.addScanParameter('V3', 80, 90, 6)
par.addUserOutput('J_R1', 'A', 'ecR4_X', 'ccJR1_R1')
# par.addUserOutput('J_R1', 'A')
par.Run()
t4 = time.time()
print("Duration: %.2f seconds" % (t4 - t3))
par.statespersecond = par.Tsteps / (t4 - t3)
print("States per second: %.1f" % par.statespersecond)

print(
    "\n Speedup with load balanced TaskClient: %.2f"
    % (par.statespersecond / ser.statespersecond)
)

print("\n\nParallel execution...using RunScatter")
par2 = pysces.ParScanner(m, engine='ipcluster')
t5 = time.time()
par2.addScanParameter('V4', 60, 100, 11)
par2.addScanParameter('V1', 100, 160, 16)
par2.addScanParameter('V2', 100, 130, 16, follower=True)
par2.addScanParameter('V3', 80, 90, 6)
par2.addUserOutput('J_R1', 'A', 'ecR4_X', 'ccJR1_R1')
# par2.addUserOutput('J_R1', 'A')
par2.RunScatter()
t6 = time.time()
print("Duration: %.2f seconds" % (t6 - t5))
par2.statespersecond = par2.Tsteps / (t6 - t5)
print("States per second: %.1f" % par2.statespersecond)

print("\n Speedup with RunScatter: %.2f" % (par2.statespersecond / ser.statespersecond))

print("\n===========\nComparing results...")
# comp = np.equal(ser.SteadyStateResults, par.SteadyStateResults[::5])
# print np.alltrue(comp)
# comp2 = np.equal(ser.UserOutputResults, par.UserOutputResults[::5])
# print np.alltrue(comp)

sss = ser.SteadyStateResults
pss = par.SteadyStateResults
suo = ser.UserOutputResults
puo = par.UserOutputResults
p2ss = par2.SteadyStateResults
p2uo = par2.UserOutputResults

print("serial vs. parallel s/s results     : ", np.alltrue(np.equal(sss, pss)))
print("serial vs. parallel user output     : ", np.alltrue(np.equal(suo, puo)))
print("TaskClient vs RunScatter s/s results: ", np.alltrue(np.equal(pss, p2ss)))
print("TaskClient vs RunScatter user output: ", np.alltrue(np.equal(puo, p2uo)))


os.chdir(backupdir)



"""
PySCeS - Python Simulator for Cellular Systems (http://pysces.sourceforge.net)

Copyright (C) 2004-2022 B.G. Olivier, J.M. Rohwer, J.-H.S Hofmeyr all rights reserved,

Brett G. Olivier (bgoli@users.sourceforge.net)
Triple-J Group for Molecular Cell Physiology
Stellenbosch University, South Africa.

Permission to use, modify, and distribute this software is given under the
terms of the PySceS (BSD style) license. See LICENSE.txt that came with
this distribution for specifics.

NO WARRANTY IS EXPRESSED OR IMPLIED.  USE AT YOUR OWN RISK.
Brett G. Olivier
"""
from __future__ import division, print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from pysces.version import __version__

__doc__ = '''network and internet oriented utilities'''

from time import strftime
from getpass import getuser


class PyscesHTML:
    """PySCeS HTML formatting class: contains some basic html elements that can be used in generated reports."""

    __version__ = __version__

    def HTML_header(self, File):
        """
        HTML_header(File)

        Write an HTML page header to file (use with HTML_footer)

        Arguments:
        =========
        File: an open, writable Python file object

        """
        header = '\n'
        header += '<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN">\n'
        header += '<html>\n'
        header += '<head>\n'
        header += (
            '<title>PySCeS data generated at '
            + strftime("%H:%M:%S (%Z)")
            + '</title>\n'
        )
        header += (
            '<meta http-equiv="Content-Type" content="text/html; charset=iso-8859-1">\n'
        )
        header += '</head>\n'
        header += '<body>\n\n'
        header += '<h4><a href="http://pysces.sourceforge.net">PySCeS</a></h4>\n\n'
        File.write(header)
        File.write('<!-- PySCeS data generated at ' + strftime("%H:%M:%S") + '-->\n\n')
        return File

    def HTML_footer(self, File):
        """
        HTML_footer(File)

        Write an HTML page footer to file (use with HTML_header)

        Arguments:
        =========
        File: an open, writable Python file object

        """
        File.write(
            '\n<p><a href="http://pysces.sourceforge.net"><font size="3">PySCeS '
            + __version__
            + '</font></a><font size="2"> output\n generated at '
            + strftime("%H:%M:%S")
            + ' by <i>'
        )
        try:
            File.write(getuser())
        except:
            File.write('PySCeS')
        File.write('</i>)</font></p>\n')
        File.write('</body>\n')
        File.write('</html>\n\n')
        return File

    def par(self, str, File=None, align='l', txtout=0):
        """
        par(str,File=None,align='l',txtout=0)

        Format <par> text and write it to a file (or string)

        Arguments:
        =========
        str: the string of text to be written
        File [default=None]: an open, writable, Python file object
        align [default='l']: HTML alignment attribute ('l','c','r')
        txtout [default=0]: do not write to file (1) return formatted HTML string

        """
        if not txtout:
            assert type(File) == file, 'The 2nd argument needs to be an open file'
        if align == 'l':
            align = 'left'
        elif align == 'r':
            align == 'right'
        elif align == 'c':
            align = 'center'
        else:
            align = ''

        strout = '\n<p align="' + align + '">'
        cntr = 0
        max_str_len = 75
        seeker_active = 0
        for x in range(len(str)):
            cntr += 1
            strout += str[x]
            if seeker_active:
                if str[x] == ' ' or str[x] == '.' or str[x] == ',':
                    cntr = max_str_len
                    seeker_active = 0
            if cntr >= max_str_len:
                if str[x] == ' ' or str[x] == '.' or str[x] == ',':
                    strout += '\n '
                else:
                    seeker_active = 1
                cntr = 0

        strout += '\n</p>\n'
        if txtout:
            return strout
        else:
            File.write(strout)
        del str
        del strout

    def h1(self, str, File=None, align='l', txtout=0):
        """
        h1(str,File=None,align='l',txtout=0)

        Format <h1> text and write it to a file (or string)

        Arguments:
        =========
        str: the string of text to be written
        File [default=None]: an open, writable, Python file object
        align [default='l']: HTML alignment attribute ('l','c','r')
        txtout [default=0]: do not write to file (1) return formatted HTML string

        """
        if not txtout:
            assert type(File) == file, 'The 2nd argument needs to be an open file'
        if align == 'l':
            align = 'left'
        elif align == 'r':
            align == 'right'
        elif align == 'c':
            align = 'center'
        else:
            align = ''

        strout = '\n<h1 align="' + align + '">'

        cntr = 0
        max_str_len = 75
        seeker_active = 0
        for x in range(len(str)):
            cntr += 1
            strout += str[x]
            if seeker_active:
                if str[x] == ' ' or str[x] == '.' or str[x] == ',':
                    cntr = max_str_len
                    seeker_active = 0
            if cntr >= max_str_len:
                print(str[x])
                if str[x] == ' ' or str[x] == '.' or str[x] == ',':
                    strout += '\n '
                else:
                    seeker_active = 1
                cntr = 0

        strout += '\n</h1>\n'
        if txtout:
            return strout
        else:
            File.write(strout)
        del str
        del strout

    def h2(self, str, File=None, align='l', txtout=0):
        """
        h2(str,File=None,align='l',txtout=0)

        Format <h2> text and write it to a file (or string)

        Arguments:
        =========
        str: the string of text to be written
        File [default=None]: an open, writable, Python file object
        align [default='l']: HTML alignment attribute ('l','c','r')
        txtout [default=0]: do not write to file (1) return formatted HTML string

        """
        if not txtout:
            assert type(File) == file, 'The 2nd argument needs to be an open file'
        if align == 'l':
            align = 'left'
        elif align == 'r':
            align == 'right'
        elif align == 'c':
            align = 'center'
        else:
            align = ''

        strout = '\n<h2 align="' + align + '">'
        cntr = 0
        max_str_len = 75
        seeker_active = 0
        for x in range(len(str)):
            cntr += 1
            strout += str[x]
            if seeker_active:
                if str[x] == ' ' or str[x] == '.' or str[x] == ',':
                    cntr = max_str_len
                    seeker_active = 0
            if cntr >= max_str_len:
                print(str[x])
                if str[x] == ' ' or str[x] == '.' or str[x] == ',':
                    strout += '\n '
                else:
                    seeker_active = 1
                cntr = 0

        strout += '\n</h2>\n'
        if txtout:
            return strout
        else:
            File.write(strout)
        del str
        del strout

    def h3(self, str, File=None, align='l', txtout=0):
        """
        h3(str,File=None,align='l',txtout=0)

        Format <h3> text and write it to a file (or string)

        Arguments:
        =========
        str: the string of text to be written
        File [default=None]: an open, writable, Python file object
        align [default='l']: HTML alignment attribute ('l','c','r')
        txtout [default=0]: do not write to file (1) return formatted HTML string

        """
        if not txtout:
            assert type(File) == file, 'The 2nd argument needs to be an open file'
        if align == 'l':
            align = 'left'
        elif align == 'r':
            align == 'right'
        elif align == 'c':
            align = 'center'
        else:
            align = ''

        strout = '\n<h3 align="' + align + '">'
        cntr = 0
        max_str_len = 75
        seeker_active = 0
        for x in range(len(str)):
            cntr += 1
            strout += str[x]
            if seeker_active:
                if str[x] == ' ' or str[x] == '.' or str[x] == ',':
                    cntr = max_str_len
                    seeker_active = 0
            if cntr >= max_str_len:
                print(str[x])
                if str[x] == ' ' or str[x] == '.' or str[x] == ',':
                    strout += '\n'
                else:
                    seeker_active = 1
                cntr = 0

        strout += '\n</h3>\n'
        if txtout:
            return strout
        else:
            File.write(strout)
        del str
        del strout


import email
import email.utils
import mimetypes

import smtplib
from email.mime.text import MIMEText

from time import sleep, strftime
from getpass import getuser

import os


class PyscesSMTP:
    """A purely experimental class that extends PySCeS with SMTP mailer capabilities. Initialise with
    sender address and local mail server name."""

    __smtp_active = 0

    def __init__(self, fromadd, server):
        self.server = server
        try:
            self.userstr = getuser()
        except:
            self.userstr = 'PySCeS '
        self.msgintro = ''
        self.fromhead = self.userstr + ' <' + fromadd + '>'
        self.signature = (
            3 * '\n'
            + '---\nSent using PySCeS 0.2.2 (http://pysces.sourceforge.net/)\n '
        )

        # auto-open connection now closed
        # self.SMTPOpen()

    def GenericMail(self, toadd, msgtxt, subj='PySCeS generated email'):
        """
        GenericMail( toadd, msgtxt, subj='PySCeS generated email')

        Generate and send a text (non-mime) email message

        Arguments:
        =========
        toadd: recipient address
        msgtxt: the message body as a string
        subj [default='PySCeS generated email']: message subject line

        """
        assert type(msgtxt) == str, '\nMessage text must be a string'
        assert self.__smtp_active, 'SMTP Server not active\n'

        msgtxt = self.msgintro + msgtxt

        msgtxt += self.signature
        outer = MIMEText(msgtxt)

        outer['Subject'] = subj
        outer['To'] = toadd
        outer['From'] = self.fromhead
        outer['Date'] = email.Utils.formatdate(localtime='true')
        outer.epilogue = ' '

        if self.CheckGo():
            try:
                self.__SMTPserver.sendmail(self.fromhead, toadd, outer.as_string())
            except SMTPServerDisconnected as e:
                print(e)
                self.SMTPOpen()
                self.__SMTPserver.sendmail(self.fromhead, toadd, outer.as_string())
            sleep(0.2)
        else:
            print('\nEmail send aborted')

    def CheckGo(self):
        """
        CheckGo()

        Do you want to continue yes or no?
        Returns 1 or 0

        Arguments:
        None

        """
        GO = 1
        while GO:
            resp = input('\nDo you want to continue (yes/no): ')
            if resp.lower() == 'yes':
                print('OK.')
                GO = 0
                return 1
            elif resp.lower() == 'no':
                print('Skipped.')
                GO = 0
                return 0
            else:
                print('\nyes to continue, no to exit')

    ##	def GenericMailHTML(self, toadd, msgtxt, htmltxt, subj='PySCeS generated email'):
    ##        """
    ##        GenericMailHTML( toadd, msgtxt, htmltxt, subj='PySCeS generated email')
    ##
    ##        Generate a mime-compliant HTML email message
    ##
    ##        Arguments:
    ##        =========
    ##        toadd: recipient address
    ##        msgtxt: text only message string
    ##        htmltxt: html formatted message string
    ##        subj [default='PySCeS generated email']: the subject line
    ##
    ##        """
    ##		assert type(msgtxt) == str, '\nMessage text must be a string'
    ##		assert self.__smtp_active, 'SMTP Server not active\n'
    ##		# Create the enclosing (outer) message
    ##		outer = email.MIMEMultipart.MIMEMultipart()
    ##		outer['Subject'] = subj
    ##		outer['To'] = toadd
    ##		outer['From'] = self.fromhead
    ##		outer['Date'] = email.Utils.formatdate(localtime='true')
    ##		outer.preamble = ' \n'
    ##		outer.epilogue = '\n---\nGenerated by PySCeS 0.2.2\n '
    ##
    ##		msgtxt += self.signature
    ##		msg = email.MIMEText.MIMEText(msgtxt)
    ##		msg.add_header('Content-Disposition', 'inline')
    ##		outer.attach(msg)
    ##
    ##		self.__SMTPserver.sendmail(self.fromhead,toadd,outer.as_string())
    ##
    ##		ctype='text/plain'
    ##		maintype, subtype = ctype.split('/', 1)
    ##		fp = open(infile, 'r')
    ##		att = email.MIMEBase.MIMEBase(maintype, subtype)
    ##		att.set_payload(fp.read())
    ##		fp.close()
    ##		# Encode the payload using Base64
    ##		#email.Encoders.encode_base64(att)
    ##		# Set the filename parameter
    ##		att.add_header('Content-Disposition', 'attachment', filename=infile)
    ##		outer.attach(att)
    ##
    ##		SMTPserver.sendmail(fromhead,toadd,outer.as_string())
    ##
    ##		sleep(0.2)      #seconds

    def SMTPOpen(self):
        """
        SMTPOpen()

        Start client and connect to an SMTP server

        Arguments:
        None

        """
        self.__SMTPserver = smtplib.SMTP(self.server)
        self.__smtp_active = 1
        print('\nSMTP server connection opened\n')

    def SMTPClose(self):
        """
        SMTPClose()

        Close connection to SMTP server

        Arguments:
        None

        """
        self.__SMTPserver.close()
        self.__smtp_active = 0
        print('\nSMTP server connection closed\n')


if __name__ == '__main__':
    replyTo = 'bgoli@sun.ac.za'
    server = 'mail.sun.ac.za'
    print('Reply to:', replyTo)
    print('SMTP server:', server)
    smtp = PyscesSMTP(replyTo, server)
    smtp.GenericMail(
        'bgoli@sun.ac.za',
        'This test message created: ' + strftime("%a, %d %b %Y %H:%M:%S"),
    )
    # smtp.GenericMail('jr@sun.ac.za','This test message created: '+ strftime("%a, %d %b %Y %H:%M:%S"))
    smtp.SMTPClose()



"""
PySCeS - Python Simulator for Cellular Systems (http://pysces.sourceforge.net)

Copyright (C) 2004-2022 B.G. Olivier, J.M. Rohwer, J.-H.S Hofmeyr all rights reserved,

Brett G. Olivier (bgoli@users.sourceforge.net)
Triple-J Group for Molecular Cell Physiology
Stellenbosch University, South Africa.

Permission to use, modify, and distribute this software is given under the
terms of the PySceS (BSD style) license. See LICENSE.txt that came with
this distribution for specifics.

NO WARRANTY IS EXPRESSED OR IMPLIED.  USE AT YOUR OWN RISK.
Brett G. Olivier
"""
from __future__ import division, print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from pysces.version import __version__

__doc__ = '''
            PyscesConfig
            ------------

            This module contains templates for the default configuration files
            on POSIX and WIN32 systems as well as utility functions for reading
            and writing them.
            '''

try:
    import configparser  # Py 3
except ImportError:
    import ConfigParser as configparser  # Py 2

import string
import os

if os.sys.platform == 'win32':
    __DefaultWin = {
        "install_dir": os.path.join(os.sys.prefix, 'lib', 'site-packages', 'pysces'),
        "model_dir": os.path.join(os.getenv('USERPROFILE'), 'Pysces', 'psc'),
        "output_dir": os.path.join(os.getenv('USERPROFILE'), 'Pysces'),
        "gnuplot_dir": None,
        "pitcon": True,
        "nleq2": True,
        "gnuplot": False,
        "matplotlib": True,
        "matplotlib_backend": 'TKagg',
        "silentstart": False,
        "change_dir_on_start": False,
    }
    __DefaultWinUsr = {
        "model_dir": os.path.join(os.getenv('USERPROFILE'), 'Pysces', 'psc'),
        "output_dir": os.path.join(os.getenv('USERPROFILE'), 'Pysces'),
        "silentstart": False,
        "change_dir_on_start": False,
    }
else:
    if hasattr(os.sys, 'lib'):
        lib = os.sys.lib
    else:
        lib = 'lib'
    __DefaultPosix = {
        "install_dir": os.path.join(
            os.sys.prefix,
            lib,
            "python%d.%d" % tuple(os.sys.version_info[:2]),
            'site-packages',
            'pysces',
        ),
        "model_dir": os.path.join(os.path.expanduser('~'), 'Pysces', 'psc'),
        "output_dir": os.path.join(os.path.expanduser('~'), 'Pysces'),
        "gnuplot_dir": None,
        "pitcon": True,
        "nleq2": True,
        "gnuplot": False,
        "matplotlib": True,
        "matplotlib_backend": 'TKagg',
        "silentstart": False,
        "change_dir_on_start": False,
    }
    __DefaultPosixUsr = {
        "model_dir": os.path.join(os.path.expanduser('~'), 'Pysces', 'psc'),
        "output_dir": os.path.join(os.path.expanduser('~'), 'Pysces'),
        "silentstart": False,
        "change_dir_on_start": False,
    }
    # OSX patch by AF
    if os.sys.platform == 'darwin':
        __DefaultPosix["matplotlib_backend"] = 'MacOSX'


def ReadConfig(file_path, config={}):
    """
    Read a PySCeS configuration file

    - *file_path*	full path to file
    - *config [default={}]* configuration data
    """
    filein = open(file_path, 'r')
    cp = configparser.ConfigParser()
    cp.readfp(filein)
    for sec in cp.sections():
        name = sec.lower()
        for opt in cp.options(sec):
            config[opt.lower()] = cp.get(sec, opt).strip()
    filein.close()
    return config


def WriteConfig(file_path, config={}, section='Pysces'):
    """
    Write a PySCeS configuration file

    - *file_path* full path to file
    - *config* [default={}]: config dictionary
    - *section* [default='Pysces']: default man section name

    """
    cfgfile = open(file_path, 'w')
    cp = configparser.ConfigParser()
    cp.add_section(section)
    for key in config:
        cp.set(section, key, str(config[key]))
    cp.write(cfgfile)
    cfgfile.close()



"""
PySCeS - Python Simulator for Cellular Systems (http://pysces.sourceforge.net)

Copyright (C) 2004-2022 B.G. Olivier, J.M. Rohwer, J.-H.S Hofmeyr all rights reserved,

Permission to use, modify, and distribute this software is given under the
terms of the PySceS (BSD style) license. See LICENSE.txt that came with
this distribution for specifics.

NO WARRANTY IS EXPRESSED OR IMPLIED.  USE AT YOUR OWN RISK.

Authors: Brett G. Olivier and Johann M. Rohwer (see CONTRIBUTORS.md for details)
"""
from __future__ import division, print_function
from __future__ import absolute_import
from __future__ import unicode_literals
from .version import __version__

__doc__ = """
            PySCeS: the Python Simulator for Cellular Systems
            -------------------------------------------------
            PySCeS is developed by the Triple-J Group for Molecular Cell Physiology
            in order to try model and understand the complex processes and systems
            which make up the living cell. PySCeS features, amongst other things:

            - A text based Model Description Language.
            - A structural analysis module.
            - Integrators for time simulation
            - Non-linear solvers for steady-state analysis
            - A module for performing Metabolic Control Analysis
            - A bifurcation module for systems which exhibit multiple steady states
            - A variety of extra utilites for parameter scans, data output and plotting.
            - A dynamic module loading framework.
            - SBML import and export capability.

            Authors: Brett G. Olivier and Johann M. Rohwer (see CONTRIBUTORS.md for details)

            Permission to use, modify, and distribute this software is given under the
            terms of the PySceS (BSD style) license. See LICENSE.txt that came with
            this distribution for specifics.

            NO WARRANTY IS EXPRESSED OR IMPLIED.  USE AT YOUR OWN RISK.
            Copyright (C) 2004-2022 B.G. Olivier, J.M. Rohwer, J.-H.S Hofmeyr all rights reserved,
            """

import os, time
from pkg_resources import get_build_platform
from . import PyscesConfig
from . import PyscesParse
from . import PyscesLink as link
from . import PyscesSED as SED

from .PyscesUtils import str2bool
from .PyscesModelMap import ModelMap

# TODO get rid unused imports
from .PyscesWeb import PyscesHTML

html = PyscesHTML()

DEBUG = False

inipath = None
lpath = None
install_dir = None
output_dir = None
model_dir = None
pitcon_switch = False
nleq2_switch = False
__USE_MATPLOTLIB__ = True
__MATPLOTLIB_BACKEND__ = 'TkAgg'
__USE_GNUPLOT__ = False
__SILENT_START__ = False

extra_dll_dir = os.path.join(os.path.dirname(__file__), '.libs')
if os.sys.platform == 'win32' and os.path.isdir(extra_dll_dir):
    os.environ["PATH"] += os.pathsep + extra_dll_dir
    if hasattr(os, 'add_dll_directory'):
        os.add_dll_directory(extra_dll_dir)

if os.sys.platform == 'win32':
    __PyscesConfigDefault = PyscesConfig.__DefaultWin
else:
    __PyscesConfigDefault = PyscesConfig.__DefaultPosix

if DEBUG:
    print(time.strftime('1-%H:%M:%S'))

eggdir = 'pysces-%s-py%s.%s-%s.egg' % (
    __version__,
    os.sys.version_info[0],
    os.sys.version_info[1],
    get_build_platform(),
)
for path in os.sys.path:
    chkPath = path.split(os.path.sep)[-1]
    if chkPath == 'pysces' and path != os.getcwd():
        if os.path.isdir(os.path.join(path, 'pysces')):
            # for in-place development with setup.py develop
            install_dir = os.path.join(path, 'pysces')
        else:
            install_dir = path
        inipath = os.path.join(install_dir, 'pyscfg.ini')
        break
    elif chkPath == eggdir:
        install_dir = os.path.join(path, 'pysces')
        inipath = os.path.join(install_dir, 'pyscfg.ini')
        break
if inipath == None:
    for k in os.sys.path_importer_cache:
        if k.split(os.path.sep)[-1] == 'pysces':
            install_dir = k
            inipath = os.path.join(install_dir, 'pyscfg.ini')
            break
        elif k.split(os.path.sep)[-1] == eggdir:
            install_dir = os.path.join(k, 'pysces')
            inipath = os.path.join(install_dir, 'pyscfg.ini')
            break
del eggdir
if DEBUG:
    print(time.strftime('2-%H:%M:%S'))

try:
    __config_dict = PyscesConfig.ReadConfig(inipath, config=__PyscesConfigDefault)
except Exception as ex:
    if DEBUG:
        print(ex)
    if not __SILENT_START__:
        print('Cwd', os.getcwd())
        print('\nWARNING: Cannot read pyscfg.ini using default values\n')
    __config_dict = __PyscesConfigDefault

# Read config
for key in __config_dict:
    if key == 'pitcon':
        pitcon_switch = str2bool(__config_dict[key])
    elif key == 'nleq2':
        nleq2_switch = str2bool(__config_dict[key])
    elif key == 'matplotlib':
        __USE_MATPLOTLIB__ = str2bool(__config_dict[key])
    elif key == 'matplotlib_backend':
        __MATPLOTLIB_BACKEND__ = __config_dict[key]
    elif key == 'gnuplot':
        __USE_GNUPLOT__ = str2bool(__config_dict[key])
    elif key == 'gnuplot_dir':
        GNUPLOT_DIR = __config_dict[key]
        if GNUPLOT_DIR == 'None':
            GNUPLOT_DIR = None
    elif key == 'silentstart':
        __SILENT_START__ = str2bool(__config_dict[key])
    elif key == 'change_dir_on_start':
        __CHGDIR_ON_START__ = str2bool(__config_dict[key])
assert inipath != None, '\nNo configuration file found'

if DEBUG:
    print(time.strftime('3-%H:%M:%S'))

__userdict = None
if os.sys.platform != 'win32':
    if os.path.exists(
        os.path.join(os.path.expanduser('~'), 'Pysces', '.pys_usercfg.ini')
    ):
        __userdict = PyscesConfig.ReadConfig(
            os.path.join(os.path.expanduser('~'), 'Pysces', '.pys_usercfg.ini'),
            PyscesConfig.__DefaultPosixUsr,
        )
    else:
        if not os.path.exists(os.path.join(os.path.expanduser('~'), 'Pysces')):
            os.makedirs(os.path.join(os.path.expanduser('~'), 'Pysces'))
        PyscesConfig.WriteConfig(
            os.path.join(os.path.expanduser('~'), 'Pysces', '.pys_usercfg.ini'),
            config=PyscesConfig.__DefaultPosixUsr,
            section='Pysces',
        )
        __userdict = PyscesConfig.ReadConfig(
            os.path.join(os.path.expanduser('~'), 'Pysces', '.pys_usercfg.ini'),
            PyscesConfig.__DefaultPosixUsr,
        )
else:
    if os.path.exists(
        os.path.join(os.getenv('HOMEDRIVE') + os.path.sep, 'Pysces', '.pys_usercfg.ini')
    ):
        __userdict = PyscesConfig.ReadConfig(
            os.path.join(
                os.getenv('HOMEDRIVE') + os.path.sep, 'Pysces', '.pys_usercfg.ini'
            ),
            PyscesConfig.__DefaultWinUsr,
        )
    elif os.path.exists(
        os.path.join(os.getenv('USERPROFILE'), 'Pysces', '.pys_usercfg.ini')
    ):
        __userdict = PyscesConfig.ReadConfig(
            os.path.join(os.getenv('USERPROFILE'), 'Pysces', '.pys_usercfg.ini'),
            PyscesConfig.__DefaultWinUsr,
        )
    else:
        if not os.path.exists(os.path.join(os.getenv('USERPROFILE'), 'Pysces')):
            os.makedirs(os.path.join(os.getenv('USERPROFILE'), 'Pysces'))
        PyscesConfig.WriteConfig(
            os.path.join(os.getenv('USERPROFILE'), 'Pysces', '.pys_usercfg.ini'),
            config=PyscesConfig.__DefaultWinUsr,
            section='Pysces',
        )
        __userdict = PyscesConfig.ReadConfig(
            os.path.join(os.getenv('USERPROFILE'), 'Pysces', '.pys_usercfg.ini'),
            PyscesConfig.__DefaultWinUsr,
        )
for key in __userdict:
    if key == 'output_dir':
        output_dir = __userdict[key]
        if not os.path.exists(__userdict[key]):
            os.makedirs(__userdict[key])
    elif key == 'model_dir':
        model_dir = __userdict[key]
        if not os.path.exists(__userdict[key]):
            os.makedirs(__userdict[key])
    elif key == 'matplotlib':
        __USE_MATPLOTLIB__ = str2bool(__userdict[key])
    elif key == 'matplotlib_backend':
        __MATPLOTLIB_BACKEND__ = __userdict[key]
    elif key == 'gnuplot':
        __USE_GNUPLOT__ = str2bool(__userdict[key])
    elif key == 'gnuplot_dir':
        GNUPLOT_DIR = __userdict[key]
        if GNUPLOT_DIR == 'None':
            GNUPLOT_DIR = None
    elif key == 'silentstart':
        __SILENT_START__ = str2bool(__userdict[key])
    elif key == 'change_dir_on_start':
        __CHGDIR_ON_START__ = str2bool(__userdict[key])
assert output_dir != None, '\nNo output directory defined'
assert model_dir != None, '\nNo model directory defined'

# following is to get the full path when .pys_usercfg.ini specifies CWD as follows:
# output_dir = ./
backup_dir = os.getcwd()
os.chdir(output_dir)
output_dir = os.getcwd()
os.chdir(backup_dir)
os.chdir(model_dir)
model_dir = os.getcwd()
os.chdir(backup_dir)

del PyscesConfig

if DEBUG:
    print(time.strftime('4-%H:%M:%S'))

# initialise pysces.interface.*
try:
    from . import PyscesInterfaces

    interface = PyscesInterfaces.Core2interfaces()
except Exception as ex:
    if DEBUG:
        print(ex)
    if not __SILENT_START__:
        print('INFO: pysces.interface.* not available')
    interface = None

# initialise pysces.plt.*
from . import PyscesPlot2

gplt = None
mplt = None
if __USE_MATPLOTLIB__:
    try:
        mplt = PyscesPlot2.MatplotlibUPI(
            work_dir=output_dir, backend=__MATPLOTLIB_BACKEND__
        )
        if not __SILENT_START__:
            print('Matplotlib interface loaded (pysces.plt.m)')
    except Exception as ex:
        if not __SILENT_START__:
            print('Matplotlib interface not available')
        if DEBUG:
            print(ex)
        __USE_MATPLOTLIB__ = False
if __USE_GNUPLOT__:
    if GNUPLOT_DIR == None or not os.path.exists(GNUPLOT_DIR):
        if not __SILENT_START__:
            print(
                '''GnuPlot has been enabled but the path to the executable has
    not been defined (or does not exist). Please set the "gnuplot_dir" key
    in your pyscfg.ini file.
            '''
            )
    else:
        try:
            if DEBUG:
                print(GNUPLOT_DIR)
            gplt = PyscesPlot2.GnuPlotUPI(work_dir=output_dir, gnuplot_dir=GNUPLOT_DIR)
            if not __SILENT_START__:
                print('GnuPlot interface loaded (pysces.plt.g)')
        except Exception as ex:
            if not __SILENT_START__:
                print('GnuPlot interface not available')
            if DEBUG:
                print(ex)
            __USE_GNUPLOT__ = False

plt = None
if __USE_MATPLOTLIB__ or __USE_GNUPLOT__:
    plt = PyscesPlot2.PyscesUPI()
if __USE_MATPLOTLIB__ and not __USE_GNUPLOT__:
    plt.p_setInterface('matplotlib', mplt)
elif __USE_GNUPLOT__ and not __USE_MATPLOTLIB__:
    plt.p_setInterface('gnuplot', gplt)
elif __USE_GNUPLOT__ and __USE_MATPLOTLIB__:
    plt.p_setInterface('matplotlib', mplt)
    plt.p_setInterface('gnuplot', gplt)
    plt.p_deactivateInterface('gnuplot')

if DEBUG:
    print(time.strftime('5-%H:%M:%S'))

alt_import = False
alt_import_pitcon = False
alt_import_nleq2 = False
if os.sys.platform == 'win32':
    if pitcon_switch:
        os.sys.path.append(os.path.join(install_dir, 'pitcon'))
        try:
            from .pitcon import pitcon as pitcon

            if not __SILENT_START__:
                print('Continuation routines available')
        except Exception as ex:
            try:
                os.environ['path'] = '%s;%s' % (
                    os.path.join(install_dir, 'win32'),
                    os.environ['path'],
                )
                from .pitcon import pitcon as pitcon

                if not __SILENT_START__:
                    print('Continuation routines available')
            except Exception as ex:
                # print 'Attempting alternate pitcon import ...'
                # alt_import = True
                # alt_import_pitcon = True
                if DEBUG:
                    print(ex)
                if not __SILENT_START__:
                    print('INFO: Pitcon import failed: continuation not available')

    if nleq2_switch:
        os.sys.path.append(os.path.join(install_dir, 'nleq2'))
        try:
            from .nleq2 import nleq2 as nleq2

            if not __SILENT_START__:
                print('NLEQ2 routines available')
        except Exception as ex:
            try:
                os.environ['path'] = '%s;%s' % (
                    os.path.join(install_dir, 'win32'),
                    os.environ['path'],
                )
                from .nleq2 import nleq2 as nleq2

                if not __SILENT_START__:
                    print('NLEQ2 routines available')
            except Exception as ex:
                # print 'Attempting alternate nleq2 import ...'
                # alt_import = True
                # alt_import_nleq2 = True
                if DEBUG:
                    print(ex)
                if not __SILENT_START__:
                    print('INFO: NLEQ2 import failed: option not available')
else:
    if pitcon_switch:
        os.sys.path.append(os.path.join(install_dir, 'pitcon'))
        try:
            from .pitcon import pitcon as pitcon

            if not __SILENT_START__:
                print('Pitcon routines available')
        except Exception as ex:
            if DEBUG:
                print(ex)
            alt_import = True
            alt_import_pitcon = True
            if not __SILENT_START__:
                print('Attempting alternate pitcon import ...')
    if nleq2_switch:
        os.sys.path.append(os.path.join(install_dir, 'nleq2'))
        try:
            from .nleq2 import nleq2 as nleq2

            if not __SILENT_START__:
                print('NLEQ2 routines available')
        except Exception as ex:
            if DEBUG:
                print(ex)
            alt_import = True
            alt_import_nleq2 = True
            if not __SILENT_START__:
                print('Attempting alternate nleq2 import ...')

if DEBUG:
    print(time.strftime('6-%H:%M:%S'))

if alt_import:
    savedir = os.getcwd()
    for tpath in os.sys.path:
        if alt_import_pitcon:
            try:
                if (
                    os.path.exists(os.path.join(tpath, 'pysces', 'pitcon'))
                    and tpath != ''
                ):
                    os.chdir(os.path.join(tpath, 'pysces', 'pitcon'))
                    from . import pitcon

                    if not __SILENT_START__:
                        print('Continuation routines available (A)')
            except Exception as ex:
                if DEBUG:
                    print(ex)
                if not __SILENT_START__:
                    print(
                        'INFO: Alternate pitcon import failed: continuation not available'
                    )
        if alt_import_nleq2:
            try:
                if (
                    os.path.exists(os.path.join(tpath, 'pysces', 'nleq2'))
                    and tpath != ''
                ):
                    os.chdir(os.path.join(tpath, 'pysces', 'nleq2'))
                    from . import nleq2

                    if not __SILENT_START__:
                        print('NLEQ2 routines available (A)')
            except Exception as ex:
                if DEBUG:
                    print(ex)
                nleq2_switch = False
                if not __SILENT_START__:
                    print('INFO: Alternate NLEQ2 import failed: option not available')
    os.chdir(savedir)

if DEBUG:
    print(time.strftime('7-%H:%M:%S'))

# check for libsbml
try:
    import libsbml as SBML

    if not __SILENT_START__:
        print("SBML support available")
except ImportError as ex:
    SBML = None
    if DEBUG:
        print(ex)
    if not __SILENT_START__:
        print("INFO: libSBML not installed, SBML support not available.")

# This has to come at the end
from .PyscesModel import PysMod as model
from .PyscesModel import ScanDataObj as ScanDataObj

PyscesModel.interface = interface
from .PyscesTest import PyscesTest as test

write = None
try:
    from .PyscesUtils import WriteOutput

    write = WriteOutput()
    del WriteOutput
except ImportError as ex:
    pass

from .PyscesScan import PITCONScanUtils, Scanner

try:
    from .RateChar import RateChar

    if not __SILENT_START__:
        print("RateChar is available")
except Exception as ex:
    RateChar = None
    if DEBUG:
        print(ex)
    # print "RateChar not available"

# ParScanner import
try:
    from .PyscesParScan import ParScanner

    if not __SILENT_START__:
        print("Parallel scanner is available")
except ImportError as ex:
    ParScanner = None
    if DEBUG:
        print(ex)
    if not __SILENT_START__:
        print("INFO: Parallel scanner not available")

if DEBUG:
    print(time.strftime('9-%H:%M:%S'))

if not __SILENT_START__:
    print('\nPySCeS environment\n******************')
    print('pysces.model_dir = ' + model_dir)
    print('pysces.output_dir = ' + output_dir)

    print('\n\n***********************************************************************')
    print(
        '* Welcome to PySCeS ('
        + __version__
        + ') - Python Simulator for Cellular Systems   *'
    )
    print('*                http://pysces.sourceforge.net                        *')
    ##  print '*                       Somewhere In Time                             *'
    print('* Copyright(C) B.G. Olivier, J.M. Rohwer, J.-H.S. Hofmeyr, 2004-2022  *')
    print('* Triple-J Group for Molecular Cell Physiology                        *')
    print('* Stellenbosch University, ZA and VU University Amsterdam, NL         *')
    print('* PySCeS is distributed under the PySCeS (BSD style) licence, see     *')
    print('* LICENCE.txt (supplied with this release) for details                *')
    ##  print '*                 ** Read about PySCeS **                             *'
    print('* Please cite PySCeS with: doi:10.1093/bioinformatics/bti046          *')
    print('***********************************************************************')

try:
    del os, key, gplt, mplt
except Exception as ex:
    print(ex)
    print('\n\nOops I did it again error ...\n\n')
if DEBUG:
    print(time.strftime('10-%H:%M:%S'))



"""
PySCeS - Python Simulator for Cellular Systems (http://pysces.sourceforge.net)

Copyright (C) 2004-2022 B.G. Olivier, J.M. Rohwer, J.-H.S Hofmeyr all rights reserved,

Brett G. Olivier (bgoli@users.sourceforge.net)
Triple-J Group for Molecular Cell Physiology
Stellenbosch University, South Africa.

Permission to use, modify, and distribute this software is given under the
terms of the PySceS (BSD style) license. See LICENSE.txt that came with
this distribution for specifics.

NO WARRANTY IS EXPRESSED OR IMPLIED.  USE AT YOUR OWN RISK.
Brett G. Olivier
"""
from __future__ import division, print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import numpy, scipy, scipy.io

class DataTools:
    """
    Data manipulation tools, requires scipy, scipy.io
    """

    def data_filter_func(row):
        print('You should overwrite this (data_filter_func) method with your own or supply a function to data_filter')
        return False

    def DataFilter(self, data, filter_func=None, report=True):
        """Takes in an array and a function which returns a boolean for filter==True"""

        if filter_func == None:
            filter_func = self.data_filter_func

        data_out = numpy.zeros(data.shape,'d')
        rr = 0
        filtered = []
        for r in range(data_out.shape[0]):
            if filter_func(data[r,:]):
                filtered.append('Filtered: %s' % data[r,:])
            else:
                data_out[rr,:] = data[r,:]
                rr += 1
        # 'remove filtered rows'
        data_out = data_out[:rr,:]
        if report:
            for a in filtered:
                print(a)
        return data_out

    def GridSortLR(self, arr):
        """
        Sort irregular unstructured grid data into monotonically rising grid data
        (where possible) Work from left column to right (uses ultra cool NumPy functions :-)
        """
        print('arr.shape:', arr.shape)
        sorted = numpy.rot90(arr)
        print('rotated.shape:', sorted.shape)
        sorted_idx = numpy.lexsort(sorted)
        print('sorted_idx.shape', sorted_idx.shape)
        sorted = sorted.take(sorted_idx, axis=-1)
        sorted = numpy.flipud(sorted)
        arr = numpy.transpose(sorted)
        print('arr.shape:', arr.shape)
        return arr


    def writeGnuPlot_3D(self, arr, fname):
        if arr.shape[0] > 5e6:
            print("warning: writeGnuPlot_3D does not work for large arrays")
        try:
            scipy.io.write_array(fname+'_result_gplt.dat', arr, precision=9, separator=' ')
        except Exception as ex:
            print('writeGnuPlot_3D exception raised')
            print(ex)

    def vtk_scalar_func(self, val):
        """
        Manipulate the scalar (or z-axis) values
        Accepts and returns a double
        """
        return val

    def writeVTK_UnstructuredGrid(self, arr, fname, scalar_func=None):
        """
        Write arr as a VTK unstructured grid to file "fname.vtk"
        arr must either be in:
            x,y,z format (in which case z is used for scalar data) or
            z,y,z,s format (s being point scalar value)
        Scalar_func(val) function to manipulate the scalar (z-axis) values

        Original version http://mayavi.sourceforge.net/mwiki/PointClouds
        Hacked quite a bit by brett 20070221
        """
        assert arr.shape[1] == 3 or arr.shape[1] == 4, '\nneed 3 or 4 columns for this'
        if scalar_func == None:
            scalar_func = self.vtk_scalar_func
        if arr.shape[1] == 4:
            HAVE_SCALARS = 1
        else:
            HAVE_SCALARS = 0
            print('No scalar values supplied Z axis values will be used')

        n=arr.shape[0]
        print("n:",n)
        # write data to vtk polydata file
        # write header
        out = open(fname+'.vtk', 'w')
        h1 = "# vtk DataFile Version 2.0\n"
        h1 += "%s\n" % fname
        h1 += "ASCII\n"
        h1 += "DATASET UNSTRUCTURED_GRID\n"
        h1 += "POINTS " + str(n) + " double\n"
        out.write(h1)
        # write xyz data
        for r in range(n):
            #s = '%15.2f %15.2f %15.2f' % (x[i], y[i], z[i])
            out.write(str(arr[r,0])+" "+str(arr[r,1])+" "+str(arr[r,2])+'\n')

        # write cell data
        out.write("CELLS "+ str(n)+ " "+ str(2*n)+'\n')
        for r in range(n):
                #s = '1 %d \n' % (i)
                out.write("1 "+str(r)+"\n")

        # write cell types
        out.write("CELL_TYPES " + str(n)+'\n')
        for r in range(n):
            out.write("1 \n")

        # write z scalar values
        h2 = '\n' + """POINT_DATA """ + str(n) + "\n"
        h3 = "SCALARS %s double 1\n" % fname
        h3 += "LOOKUP_TABLE default\n"
        out.write(h2 + h3)

        for r in range(n):
            if HAVE_SCALARS:
                sc=(scalar_func(arr[r,3]))
            else:
                sc=(scalar_func(arr[r,2]))
            out.write(str(sc)+ "\n")

        out.write('\n')
        out.close()

class IntersectionAnalysis(DataTools):
    zmin = 0.01
    intersect_tol = 0.04
    coords = None
    vtkcoords = None
    intersection = None
    vtkintersection = None

    def setCoords(self, coords):
        self.coords = coords

    def setVTKCoords(self):
        self.vtkcoords = numpy.zeros((self.coords.shape[0], self.coords.shape[1]+1),'d')
        self.vtkcoords[:,:self.coords.shape[1]] = self.coords.copy()

    def data_filter_func(self, row):
        val = False
        ##  if (numpy.abs(row) < 1.0e-10).any() or (row < 0.0).any():
        if (row < self.zmin).any():
            val = True
        return val

    def CalculateIntersect(self, tol=None):
        """calc and add scalar value for intersection to vtkcoords"""
        if tol != None:
            self.intersect_tol = tol
        intersect = []
        cntr = 0
        for r in range(self.coords.shape[0]):
            if numpy.abs(self.coords[r, 2] - self.coords[r, 3]) < self.intersect_tol:
                intersect.append(self.coords[r,:])
                cntr += 1
        print('\n*****\n%s intersections found.\n*****\n' % cntr)
        self.intersection = numpy.array(intersect)
        self.vtkintersection = self.intersection.copy()
        del intersect

    def LogVTKdata(self):
        print("Logging VTK coords")
        # log for vtk
        for r in range(self.vtkcoords.shape[0]):
            self.vtkcoords[r,:self.coords.shape[1]] = scipy.log10(self.vtkcoords[r,:self.coords.shape[1]])

    def LogVTKIntersectiondata(self):
        print("Logging VTK intersection")
        # log for vtk intersect
        for r in range(self.vtkintersection.shape[0]):
            self.vtkintersection[r,:] = scipy.log10(self.intersection[r,:])
